from transformers import PreTrainedModel, AutoModel

def get_encoder_from_model(model:PreTrainedModel):
    assert isinstance(model, PreTrainedModel), "Expected a PreTrainedModel but got %s" % type(model)
    # get encoder model class   
    model_class = AutoModel._model_mapping.get(type(model.config), None)
    assert model_class is not None, "Model type not registered!"
    # find member of encoder class in model
    for module in model.children():
        if isinstance(module, model_class):
            return module
    # attribute error
    raise AttributeError("Encoder member of class %s not found" % model_class.__name__)



import torch
from torch.utils.data import Dataset, Subset
from .strategies.strategy import AbstractStrategy
from .strategies.random import Random
from itertools import filterfalse
from typing import Sequence
from math import ceil

class ActiveLoop(object):

    def __init__(
        self,
        pool:Dataset,
        batch_size:int,
        query_size:int,
        strategy:AbstractStrategy,
        init_strategy:AbstractStrategy =Random()
    ) -> None:
        self.pool = Subset(
            dataset=pool,
            indices=list(range(len(pool)))
        )
        self.batch_size = batch_size
        self.query_size = query_size
        self.strategy = strategy
        self.init_strategy = init_strategy

    def update(self, indices:Sequence[int]) -> Subset:
        # get the data points and remove them from the pool
        data = Subset(
            dataset=self.pool.dataset,
            indices=list(indices)
        )
        self.pool.indices = list(filterfalse(
            set(indices).__contains__,
            self.pool.indices
        ))
        # return data
        return data

    def apply_strategy(self, strategy:AbstractStrategy) -> Subset:
        """ """
        # select samples using strategy
        indices = strategy.query(
            pool=self.pool,
            query_size=min(self.query_size, len(self.pool)),
            batch_size=self.batch_size
        )
        return self.update([self.pool.indices[i] for i in indices])

    def init_step(self) -> Subset:
        return self.apply_strategy(self.init_strategy)
    
    def step(self) -> Subset:
        return self.apply_strategy(self.strategy)

    def __len__(self) -> int:
        return ceil(len(self.pool) / self.query_size)

    def __iter__(self):
        # apply initial heuristic
        yield self.init_step()
        # apply heuristic unitl pool is empty
        while len(self.pool) > 0:
            yield self.step()



import torch
from torch.utils.data import Dataset, ConcatDataset, DataLoader, random_split
# import ignite
from ignite.engine import Engine
from ignite.engine.events import State, Events, EventEnum
# others
from active.core.loop import ActiveLoop
from .trainer import Trainer
from typing import Optional

class ActiveLearningEvents(EventEnum):
    """ Costum Events fired by Active Learning Engine.

        This list of events also contains some convergence
        management events. I.e. a model might be trained
        for multiple tries until the trainers convergence
        criteria are met.

        Events:
            DATA_SAMPLING_COMPLETE: 
                called after data is sampled and split into training and validation.
                The `engine.training_data` and `engine.validation_data` are up to date.
    """
    DATA_SAMPLING_COMPLETED = "data-sampling-completed"

class ActiveLearningEngine(Engine):
    """ Active Learning Engine implementing the basic active
        learning procedure of iterating the following steps:
            1. gathering the next samples from the active loop
               provided in the `run` method
            2. splitting samples into training and validation data
            3. running trainer on the full dataset sampled up to this point

        Args:
            trainer (Trainer): model trainer
            trainer_run_kwargs (Optional[dict]): 
                keyword arguments passed to the trainers run method.
                Sets values like `max_epochs` and `epoch_length`.
            train_batch_size (Optional[int]): 
                batch size used for model training.
                Defaults to 32.
            eval_batch_size (Optional[int]): 
                batch size used for model evaluation. 
                Defaults to `train_batch_size`.
            train_val_ratio (Optional[float]):
                ratio between training and validation data sizes.
                Defaults to 0.9 meaning 90% of the sampled data is
                used for model training and the remaining 10% is
                used as validation data.
    """

    def __init__(
        self,
        trainer:Trainer,
        trainer_run_kwargs:Optional[dict] ={},
        train_batch_size:Optional[int] =32,
        eval_batch_size:Optional[int] =None,
        train_val_ratio:Optional[float] =0.9
    ) -> None:
        # initialize engine
        super(ActiveLearningEngine, self).__init__(type(self).step)

        # register costum events
        self.register_events(*ActiveLearningEvents)

        # save trainer and other arguments
        self.trainer = trainer
        self.train_batch_size = train_batch_size
        self.eval_batch_size = eval_batch_size or self.train_batch_size
        self.trainer_run_kwargs = trainer_run_kwargs
        self.train_val_ratio = train_val_ratio

        # active datasets
        self.train_data = []
        self.val_data = []

        # reset engine state and datasets on start
        self.add_event_handler(Events.STARTED, type(self)._reset)

    @property
    def num_train_samples(self) -> int:
        return sum(map(len, self.train_data))

    @property
    def num_val_samples(self) -> int:
        return sum(map(len, self.val_data))

    @property
    def num_total_samples(self) -> int:
        return self.num_train_samples + self.num_val_samples

    @property
    def train_dataset(self) -> ConcatDataset:
        """ Training dataset """
        return ConcatDataset(self.train_data)
    
    @property
    def val_dataset(self) -> ConcatDataset:
        """ Validation dataset. Fallback to train dataset if no validation
            data is present yet.
        """
        if self.num_val_samples > 0:
            return ConcatDataset(self.val_data)
        else:
            return self.train_dataset

    def dataloader(self, data:Dataset, **kwargs) -> DataLoader:
        """ Create the dataloader for a given dataset with some specific configuration.            

            Args:
                data (Dataset): dataset to use
                **kwargs (Any): keyword arguments passed to the dataloader

            Returns:
                loader (DataLoader): dataloader from given dataset and configuration
        """
        return DataLoader(data, **kwargs)

    def _reset(self):
        """ Event handler to reset the engine, i.e. re-initialize the model,
            reset optimizer and scheduler states and clear the sampled datasets. 
            Called on `STARTED`.
        """
        # re-initialize model and reset optimizer and scheduler states
        self.trainer._load_init_ckpt(force=True)
        # clear datasets
        self.train_data.clear()
        self.val_data.clear()

    def _reset_state(self):
        """ Reset the internal state of the engine. Called at the very
            beginning of the `run` method.
        """
        self.state = State()

    def step(self, samples:Dataset):
        """ Engines step function implementing a single step of
            the default active learning procedure 

            Args:
                samples (Dataset): samples selected for the active learning step
        """
        # make sure samples are valid
        assert len(samples) > 0, "No samples provided!"
        # compute data split sizes satifying the train-validation ratio as closely as possible
        n_train = round((self.num_total_samples + len(samples)) * self.train_val_ratio) - self.num_train_samples
        n_val = len(samples) - n_train
        # split into train and validation samples
        train_samples, val_samples = random_split(samples, [n_train, n_val])
        
        # add to the respective datasets
        self.train_data.append(train_samples)
        self.val_data.append(val_samples)

        # fire data generated event
        self.fire_event(ActiveLearningEvents.DATA_SAMPLING_COMPLETED)
        
        # create dataloaders and update validation loader in trainer
        train_loader = self.dataloader(self.train_dataset, batch_size=self.train_batch_size, shuffle=True)
        self.trainer.val_loader = self.dataloader(self.val_dataset, batch_size=self.eval_batch_size, shuffle=False)
        # run training
        self.trainer.run(train_loader, **self.trainer_run_kwargs)

    def run(
        self, 
        loop:ActiveLoop, 
        steps:Optional[int] =None,
        seed:Optional[int] =None
    ) -> State:
        # fully reset state
        self._reset_state()
        # run
        return super(ActiveLearningEngine, self).run(
            data=loop,
            max_epochs=1,
            epoch_length=steps,
            seed=seed
        )



import torch
import heapq

from dataclasses import dataclass, field

__all__ = ["topk_sequences"]

@dataclass
class Node:
    """ Helper class """
    probs:torch.FloatTensor
    seq:tuple
    _swapped_idx:set =field(default_factory=set)

    def __post_init__(self):
        self.seq = tuple(self.seq)
        assert len(self.seq) == self.probs.size(0)

    @property
    def class_probs(self) -> torch.FloatTensor:
        return self.probs[range(len(self.seq)), self.seq]

    @property
    def posterior(self) -> float:
       return self.class_probs.prod() 

    def __lt__(self, other):
        "Inverted to turn min-heap into max-heap"
        return self.posterior > other.posterior

    def __ge__(self, other):
        return self.posterior >= other.posterior
    def __le__(self, other):
        return self.posterior <= other.posterior

    """ needed for set operations """
    def __eq__(self, other):
        return self.seq == other.seq
    def __ne__(self, other):
        return not (self == other)
    def __hash__(self):
        return hash(self.seq)

    @property
    def children(self):
        
        idx = set(tuple(range(len(self.seq))))
        for i in (idx - self._swapped_idx):
            for j in range(self.probs.size(1)):
                if j != self.seq[i]:
                    yield Node(
                        probs=self.probs,
                        seq=self.seq[:i] + (j,) + self.seq[i+1:],
                        _swapped_idx=self._swapped_idx.union({i})
                    )

    def __str__(self) -> str:
        return "(%s)" % ', '.join([str(i) for i in self.seq])

    def __repr__(self) -> str:
        return "%s with P=%.02f" % (str(self), self.posterior)


def topk_sequences(probs, k):
    """ Find the top-k sequences with highest a-posteriori probabilities.

        Args:
            probs (torch.FloatTensor):
                input tensor of sequence with specific item probabilities.
                Shape must be (S, C) where S is the sequence length and C is
                the number of classes.
            k (int): number of sequences to generate. k is upper bounded by C^S.
        
        Returns:
            probs (torch.FloatTensor): posterior probabilities of the generated sequences.
            args (torch.LongTensor):
                 argument tensor holding the top-k sequences. Shape matches (k, S).
    """
    # check input
    k = min(k, probs.size(1) ** probs.size(0))

    # build root node
    root = Node(
        probs=probs,
        seq=torch.argmax(probs, dim=1).tolist()
    )

    # list of all leave nodes
    leaves = [root]
    nodes = []
    
    # only do k-1 to avoid building children in last iteration
    for _ in range(k - 1):
        # get leave with maximum posterior
        heapq.heapify(leaves)
        node = heapq.heappop(leaves)
       
        # 
        nodes.append(node)
        # add new leaves and uniquify
        leaves = list(set(leaves + list(node.children)) - set(nodes))

    # add last node
    heapq.heapify(leaves)
    nodes.append(heapq.heappop(leaves))

    # build output
    seq = torch.LongTensor([node.seq for node in nodes])
    p = torch.FloatTensor([node.posterior for node in nodes])

    return p, seq



import torch
from torch.utils.data import Dataset
from .strategy import AbstractStrategy
from typing import Sequence

class Random(AbstractStrategy):

    def process(self, batch):
        pass

    def sample(self, output, num_samples):
        pass

    def query(
        self,
        pool:Dataset,
        query_size:int,
        batch_size:int
    ) -> Sequence[int]:
        # sample random indices within the pool
        perm = torch.randperm(len(pool))
        return perm[:query_size]



import os
import torch
import numpy as np
from torch.utils.data import Dataset, DataLoader
# import active learning components
from active.core import strategies
from active.core.loop import ActiveLoop
from active.core.metrics import AreaUnderLearningCurve, WorkSavedOverSampling
from active.helpers.engine import ActiveLearningEvents, ActiveLearningEngine
from active.helpers.evaluator import Evaluator
from active.core.utils.model import get_encoder_from_model
# import ignite
from ignite.engine import Events
from ignite.metrics import ConfusionMatrix
from ignite.contrib.handlers.tqdm_logger import ProgressBar
# dimensionality reduction
from sklearn.manifold import TSNE
# others
import wandb
import transformers
from math import ceil
from pydantic import BaseModel, validator
from matplotlib import pyplot as plt
# base config and helper function
from active.scripts.run_train import Task, ExperimentConfig, attach_metrics

def visualize_embeds(strategy):
    # get selected indices and processing output of unlabeled pool
    idx = strategy.selected_indices
    output = strategy.output
    output = output if output.ndim == 2 else \
        output.reshape(-1, 1) if output.ndim == 1 else \
        output.flatten(start_dim=1)
    # reduce dimension for visualization using t-sne    
    X = TSNE(
        n_components=2,
        perplexity=50,
        learning_rate='auto',
        init='random'
    ).fit_transform(
        X=output
    )
    # plot
    fig, ax = plt.subplots()
    ax.scatter(X[:, 0], X[:, 1], s=1.0, color="blue", alpha=0.1)
    ax.scatter(X[idx, 0], X[idx, 1], s=2.0, color='red', alpha=1.0)
    ax.legend(['pool', 'query'])
    # return axes and figure
    return ax, fig

class ActiveLearningConfig(BaseModel):
    """ Active Learning Configuration Model """
    task:Task =None
    # strategy to use
    strategy:str
    # budget and query size
    budget:int
    query_size:int

    @property
    def num_steps(self):
        return ceil(self.budget / self.query_size)

    def build_strategy(self, model:transformers.PreTrainedModel) -> strategies.AbstractStrategy:    
        if self.strategy == 'random':
            return strategies.Random()
        elif self.strategy == 'least-confidence':
            return strategies.LeastConfidence(model)
        elif self.strategy == 'prediction-entropy':
            return strategies.PredictionEntropy(model)
        elif self.strategy == 'badge' and self.task == Task.SEQUENCE:
            return strategies.BadgeForSequenceClassification(
                get_encoder_from_model(model), 
                model.classifier
            )
        elif self.strategy == 'badge' and self.task in [Task.BIO_TAGGING, Task.NESTED_BIO_TAGGING]:
            return strategies.BadgeForTokenClassification(
                get_encoder_from_model(model), 
                model.classifier
            )
        elif self.strategy == 'alps':
            return strategies.AlpsConstantEmbeddings(model, mlm_prob=0.15)
        elif self.strategy == 'egl':
            return strategies.EglByTopK(model, k=5)
        elif self.strategy == 'egl-sampling':
            return strategies.EglBySampling(model, k=8)
        elif self.strategy == 'entropy-over-max':
            return strategies.EntropyOverMax(model)
        elif self.strategy == 'entropy-over-max-ignore':
            return strategies.EntropyOverMax(model, ignore_labels=[0])
        elif self.strategy == 'entropy-over-max-sample':
            return strategies.EntropyOverMax(model, random_sample=True)

        raise ValueError("Unrecognized strategy: %s" % self.strategy)

class AlExperimentConfig(ExperimentConfig):
    """ Active Learning Experiment Configuration """

    # add active learning config to experiment config
    active:ActiveLearningConfig

    @validator('active', pre=True)
    def _pass_task_to_active_config(cls, v, values):
        assert 'task' in values
        if isinstance(v, BaseModel):
            return v.copy(update={'task': values.get('task')})
        elif isinstance(v, dict):
            return v | {'task': values.get('task')}

def main():

    from argparse import ArgumentParser
    # build argument parser
    parser = ArgumentParser(description="Train transformer model on sequence or bio-taging classification tasks using active learning.")
    # add arguments
    parser = ArgumentParser(description="Train transformer model on sequence or token classification tasks.")
    parser.add_argument("--config", type=str, required=True, help="Path to a valid experiment configuration")
    parser.add_argument("--use-cache", action='store_true', help="Load cached preprocessed datasets if available")
    parser.add_argument("--seed", type=int, default=1337, help="Random Seed")
    parser.add_argument("--strategy", type=str, default=None, help="Overwrite strategy specified in config")
    parser.add_argument("--budget", type=int, default=None, help="Overwrite budget specified in config")
    parser.add_argument("--query-size", type=int, default=None, help="Overwrite query size specified in config")
    # parse arguments
    args = parser.parse_args()

    # set random seed
    torch.manual_seed(args.seed)
    np.random.seed(args.seed)

    # parse configuration and apply overwrites
    config = AlExperimentConfig.parse_file(args.config)
    config.active.strategy = args.strategy or config.active.strategy
    config.active.budget = args.budget or config.active.budget
    config.active.query_size = args.query_size or config.active.query_size
    print("Config:", config.json(indent=2))

    # load datasets
    ds = config.load_dataset(use_cache=args.use_cache)
    
    # keep tokenizer around as it is used in event handler
    tokenizer = config.model.tokenizer
    # load model and build strategy
    model = config.load_model()
    strategy = config.active.build_strategy(model)
    # attach progress bar to strategy
    ProgressBar(ascii=True, desc='Strategy').attach(strategy)
    
    # create trainer, validator and tester engines
    trainer = config.trainer.build_trainer(model) 
    validator = Evaluator(model)
    tester = Evaluator(model)
    # attach metrics
    attach_metrics(trainer, tag="train")
    attach_metrics(validator, tag="val")
    attach_metrics(tester, tag="test")
    # attach progress bar
    ProgressBar(ascii=True).attach(trainer, output_transform=lambda output: {'L': output['loss']})
    ProgressBar(ascii=True, desc='Validating').attach(validator)
    ProgressBar(ascii=True, desc='Testing').attach(tester)
    # attach confusion matrix metric to tester
    # needed for some active learning metrics
    ConfusionMatrix(
        # for nested bio tagging the label space is the entity
        # types, however the confusion matrix is computed over B,I,O tags
        num_classes=3 if config.task is Task.NESTED_BIO_TAGGING else len(config.data.label_space),
        output_transform=tester.get_logits_and_labels
    ).attach(tester, "cm")

    # create active learning loop
    loop = ActiveLoop(
        pool=ds['train'],
        strategy=strategy,
        batch_size=config.trainer.batch_size,
        query_size=config.active.query_size,
        init_strategy=strategy if isinstance(strategy, strategies.Alps) else strategies.Random()
    )
    # create active learning engine
    al_engine = ActiveLearningEngine(
        trainer=trainer,
        trainer_run_kwargs=config.trainer.run_kwargs,
        train_batch_size=config.trainer.batch_size,
        eval_batch_size=config.trainer.batch_size,
        train_val_ratio=0.9
    )
    
    @al_engine.on(Events.ITERATION_STARTED)
    def on_started(engine):
        # log active learning step
        i = engine.state.iteration
        print("-" * 8, "AL Step %i" % i, "-" * 8)
        
    @al_engine.on(ActiveLearningEvents.DATA_SAMPLING_COMPLETED)
    def save_selected_samples(engine):
        # create path to save samples in
        path = os.path.join(
            "output",
            os.environ.get("WANDB_RUN_GROUP", "default"),
            os.environ.get("WANDB_NAME", "default-run"),
        )
        os.makedirs(path, exist_ok=True)
        # get selected samples and re-create input texts
        data = engine.state.batch
        texts = tokenizer.batch_decode(
            sequences=[sample['input_ids'] for sample in data],
            skip_special_tokens=True
        )
        # save selected samples to file
        with open(os.path.join(path, "step-%i.txt" % len(engine.train_dataset)), 'w+') as f:
            f.write('\n'.join(texts))

    @al_engine.on(Events.ITERATION_COMPLETED)
    def evaluate_and_log(engine):
        # print
        print(
            "Training Converged:", trainer.converged, 
            "Train F-Score: %.03f" % trainer.state.metrics['train/F']
        )
        # create validation and test dataloaders
        val_loader = DataLoader(engine.val_dataset, batch_size=64, shuffle=False)
        test_loader = DataLoader(ds['test'], batch_size=64, shuffle=False)
        # run on validation data
        val_metrics = validator.run(val_loader).metrics
        print("Validation Metrics:", val_metrics)

        # run on test data
        test_metrics = tester.run(test_loader).metrics
        # don't log test confusion matrix
        test_metrics = test_metrics.copy()
        test_metrics.pop('cm')
        # print test metrics
        print("Test Metrics:", test_metrics)

        # get total time spend in strategy
        strategy_time = loop.strategy.state.times[Events.COMPLETED.name]
        strategy_time = {'times/strategy': strategy_time} if strategy_time is not None else {}
        # log all remaining metrics to weights and biases
        wandb.log(
            data=(trainer.state.metrics | val_metrics | test_metrics | strategy_time),
            step=len(engine.train_dataset)
        )
    
        # check if there is an output to visualize
        if loop.strategy.output is not None:
            ax, fig = visualize_embeds(loop.strategy)
            ax.set(title="%s embedding iteration %i (t-SNE)" % (args.strategy, engine.state.iteration))
            wandb.log({"Embedding": wandb.Image(fig)}, step=len(engine.train_dataset))
            # close figure
            plt.close(fig)
    
    # add metrics to active learning engine
    wss = WorkSavedOverSampling(output_transform=lambda _: tester.state.metrics['cm'])
    area = AreaUnderLearningCurve(
        output_transform=lambda _: (
            # point of learning curve given
            # by iteration and accuracy value
            al_engine.state.iteration,
            tester.state.metrics['test/F']
        )
    )
    wss.attach(al_engine, "test/wss")
    area.attach(al_engine, "test/Area(F)")

    # initialize wandb
    wandb_config = config.dict()
    wandb_config['data']['dataset'] = config.data.dataset_info.builder_name
    wandb_config['seed'] = args.seed
    wandb.init(config=wandb_config)

    # run active learning experiment
    state = al_engine.run(loop, steps=config.active.num_steps)
    print("Active Learning Metrics:", state.metrics)
    # log active learning metric scores
    wandb.run.summary.update(state.metrics)

    # run finished
    wandb.finish(quiet=True)

if __name__ == '__main__':
    main()



import torch
import torch.nn as nn
import torch.nn.functional as F
import ignite.distributed as idist
# base strategy and utils
from .strategy import ScoreBasedStrategy
from ..utils.data import move_to_device
from ..utils.gradnorm import GoodfellowGradientNorm
from ..utils.sequence import topk_sequences
# import transformers and others
from transformers import PreTrainedModel
from abc import abstractmethod
from typing import Tuple, Sequence, Any, Union, Optional

__all__ = [
    "EglByTopK",
    "EglBySampling"
]

class _EglBase(ScoreBasedStrategy):
    """ Abstract Base for Maximum Expected Gradient Length Sampling 
        Strategy. Assumes cross-entropy loss. Child classes must
        overwrite the abstract method `_get_hallucinated_labels`
        which generates the labels for gradient computations.
    
        Args:
            model (PreTrainedModel): 
                transformer-based sequence classification model
            random_sample (Optional[bool]): 
                whether to random sample query according to distribution given by expected gradient length.
                Defaults to False, meaning the elements with maximum expected gradient langth are selected.
    """

    def __init__(
        self,
        model:PreTrainedModel,
        random_sample:Optional[bool] =False
    ) -> None:
        # initialize strategy
        super(_EglBase, self).__init__(
            random_sample=random_sample
        )
        # move model to available device(s)
        self.model = idist.auto_model(model)

    @abstractmethod
    def _get_hallucinated_labels(
        self, 
        logits:torch.FloatTensor,
        mask:Union[torch.BoolTensor, None]
    ) -> Tuple[torch.FloatTensor, torch.LongTensor]:
        """ Get hallucinated labels for which to compute the loss and gradients.
        
            Args:
                logits (torch.FloatTensor): 
                    logits of shape (n, m) where n is the number of examples 
                    and m is the number of labels. For Token Classification tasks
                    the shape is (n, s, m) where s is the sequence length.
                mask (Union[torch.BoolTensor, None]):
                    attention mask of shape (n, s, m). None if no attention mask
                    is provided by the batch.

            Returns:
                probs (torch.FloatTensor):
                    probabilities of the specific label infered from the logits.
                    Must be of shape (n, k) where k is the number of hallucinated
                    labels to choose.
                labels (torch.LongTensor):
                    hallucinated labels selected for loss computation. Must be of the
                    shape (n, k). Again k refers to the number of hallucinated labels.
                    For Token Classification tasks the shape must be (n, s, k).
        """
        raise NotImplementedError()
    
    def process(self, batch:Any) -> torch.FloatTensor:
        """ Compute expected gradient lengths for given batch 
        
            Args:
                batch (Any): input batch for which to compute grad norms
        
            Returns:
                norm (torch.FloatTensor): (squared) gradient norm length computed following Goodfellow (2016)
        """
        self.model.eval()
        # create gradient norm tracking context
        with GoodfellowGradientNorm(self.model) as gradnorm:
            # move batch to device and pass through model
            batch = move_to_device(batch, device=idist.device())
            logits = self.model(**batch).logits
            # get top-k predictions
            mask = batch.get('attention_mask', None)
            probs, labels = self._get_hallucinated_labels(logits, mask)
            probs, labels = probs.to(idist.device()), labels.to(idist.device())
            # check dimensions
            assert probs.size(0) == labels.size(0) == logits.size(0), \
                "Mismatch between batchsizes of probs (%i), labels (%i) and logits (%i)" % \
                    (probs.size(0), labels.size(0), logits.size(0))
            assert probs.size(-1) == labels.size(-1), \
                 "Mismatch between selected labels of probs (%i) and labels (%i)" % \
                 (probs.size(-1), labels.size(-1))
            assert probs.ndim == 2, \
                "Expected probabilities two have two dimensions but got shape $s" % \
                str(tuple(probs.size()))
            # compute loss and backpropagate
            for i in range(labels.size(-1)):
                F.cross_entropy(
                    logits.flatten(end_dim=-2), 
                    labels[..., i].flatten(), 
                    reduction='sum'
                ).backward(retain_graph=i < labels.size(-1)-1)
            # approximate expected gradient length
            with torch.no_grad():
                return torch.sum(probs * gradnorm.compute(), dim=1)


class EglByTopK(_EglBase):
    """ Implements Maximum Expected Gradient Length Sampling Strategy
        by selecting the top-k predictions as labels for gradient 
        computation. Assumes cross-entropy loss.
    
        Args:
            model (PreTrainedModel): 
                transformer-based sequence classification model
            k (int):
                maximum number of targets to consider for
                expectation approximation. Defaults to 5.
    """
    
    def __init__(
        self,
        model:PreTrainedModel,
        k:int =5
    ) -> None:
        # initialize strategy
        super(EglByTopK, self).__init__(model)
        self.k = k   
 
    @torch.no_grad()
    def _get_hallucinated_labels(
        self, 
        logits:torch.FloatTensor, 
        mask:Union[torch.BoolTensor, None]
    ) -> Tuple[torch.FloatTensor, torch.LongTensor]:
        """ Get hallucinated labels for which to compute the loss and gradients.
            Selects the top-k predictions as hallucinated labels.
        
            Args:
                logits (torch.FloatTensor): 
                    logits of shape (n, m) where n is the number of examples 
                    and m is the number of labels. For Token Classification tasks
                    the shape is (n, s, m) where s is the sequence length.
                mask (Union[torch.BoolTensor, None]):
                    attention mask of shape (n, s, m). None if no attention mask
                    is provided by the batch.

            Returns:
                probs (torch.FloatTensor):
                    probabilities of the specific label infered from the logits.
                    Must be of shape (n, k) where k is the number of hallucinated
                    labels to choose.
                labels (torch.LongTensor):
                    hallucinated labels selected for loss computation. Must be of the
                    shape (n, k). Again k refers to the number of hallucinated labels.
                    For Token Classification tasks the shape must be (n, s, k).
        """
        probs = torch.softmax(logits, dim=-1) 
        
        if logits.ndim == 2:
            # sequence classification
            k = min(self.k, logits.size(-1))
            return torch.topk(probs, k=k, dim=-1)
        
        if logits.ndim == 3:
            # token classification
            n, s, _ = logits.size()
            lengths = torch.full(n, fill_value=s) if mask is None else mask.sum(dim=1)
            # put probs to cpu
            probs = probs.cpu()
            # get top-k sequences
            posteriors = torch.zeros((n, self.k), dtype=float)
            labels = torch.zeros((n, s, self.k), dtype=int)
            # build top-k sequences for batch one by one
            # this is very slow but not easily vectorizable
            for i, j in enumerate(lengths):
                # get valid sub-matrix for faster top-k compuations
                p, l = topk_sequences(probs[i, :j, :], k=self.k)
                # handle less than k sequence generated due
                # note that posteriors are initialized to zeros
                # meaning invalid sequences are irrelevant
                k_ = p.size(0)                
                posteriors[i, :k_] = p
                labels[i, :j, :k_] = l.t()
            # return probabilities and labels
            return posteriors, labels
        
        raise NotImplementedError()

class EglBySampling(_EglBase):
    """ Implements Maximum Expected Gradient Length Sampling Strategy
        by random sampling labels from the predicted label distribution.
        Assumes cross-entropy loss.
    
        Args:
            model (PreTrainedModel): 
                transformer-based sequence classification model
            k (int):
                maximum number of targets to consider for
                expectation approximation. Defaults to 5.
    """
    
    def __init__(
        self,
        model:PreTrainedModel,
        k:int =5
    ) -> None:
        # initialize strategy
        super(EglBySampling, self).__init__(model)
        self.k = k   

    @torch.no_grad()
    def _get_hallucinated_labels(
        self, 
        logits:torch.FloatTensor, 
        mask:Union[torch.BoolTensor, None]
    ) -> Tuple[torch.FloatTensor, torch.LongTensor]:
        """ Get hallucinated labels for which to compute the loss and gradients.
            Samples labels from distribution defined by logits.
            
            Args:
                logits (torch.FloatTensor): 
                    logits of shape (n, m) where n is the number of examples 
                    and m is the number of labels. For Token Classification tasks
                    the shape is (n, s, m) where s is the sequence length.
                mask (Union[torch.BoolTensor, None]):
                    attention mask of shape (n, s, m). None if no attention mask
                    is provided by the batch.

            Returns:
                probs (torch.FloatTensor):
                    probabilities of the specific label infered from the logits.
                    Must be of shape (n, k) where k is the number of hallucinated
                    labels to choose.
                labels (torch.LongTensor):
                    hallucinated labels selected for loss computation. Must be of the
                    shape (n, k). Again k refers to the number of hallucinated labels.
                    For Token Classification tasks the shape must be (n, s, k).
        """
        # compute probabilities
        probs = torch.softmax(logits, dim=-1) 
        # sample labels from logits
        labels = torch.multinomial(
            probs.flatten(end_dim=-2),
            num_samples=self.k,
            replacement=True
        ).reshape(*logits.size()[:-1], self.k)
        # token classification, apply mask to disable loss for invalid tokens
        if (mask is not None) and (logits.ndim == 3):
            labels.masked_fill_(~mask.bool().unsqueeze(-1), -100)
        # weight each label the same, weighting is done by sampling
        return torch.ones((logits.size(0), self.k), device=labels.device), labels



from __future__ import annotations
import gym
import torch
import numpy as np
# ignite
from ignite.engine import Events
from ignite.metrics import Metric
import ignite.distributed as idist
# import active learning components
from active.core.loop import ActiveLoop
from active.core.strategies import AbstractStrategy, Random
from active.helpers.engine import ActiveLearningEngine
from active.helpers.evaluator import Evaluator
# import data utilities
from active.core.utils.data import (
    NamedTensorDataset,
    default_collate_drop_labels
)
from torch.utils.data import (
    Subset, 
    Dataset,
    DataLoader,
    default_collate
)
# others
from itertools import chain
from dataclasses import dataclass, field
from typing import List, Optional

@dataclass
class State:
    """ Helper class storing the internal state of a `StreamBasedEnv` instance.

        Attributes:
            budget (int): total annotation budget
            query_size (int): number of queries to sample before retraining the model

            prev_metric (float): 
                value the reward metric evaluated to in the previous active learning step.
    """
    budget:int
    query_size:int

    # current query and corresponding observation
    # I.e. the query is an element extracted from the
    # model_queue and the observation is the corresponding
    # element in the observation queue
    query:dict =None
    obs:dict =None

    # model and observation queues
    model_queue:List[dict] = field(default_factory=list)    
    obs_queue:List[dict] =field(default_factory=list)

    # selected samples for current active learning step
    # and a counter to keep track of the total number of
    # samples selected
    samples:List[int] =field(default_factory=list)
    total_samples:int =0    

    # reward metric value evaluated at the previous AL step
    prev_metric:float =0

    def reset(self) -> None:
        self.__init__(
            budget=self.budget, 
            query_size=self.query_size
        )

    @property
    def are_queues_empty(self) -> bool:
        return len(self.obs_queue) == 0

    def next_query(self) -> dict:
        # make sure queues are not empty        
        assert not self.are_queues_empty, "Queues are empty!"
        # pop first element from both queues
        self.query = self.model_queue.pop()
        self.obs = self.obs_queue.pop()
        # return observation
        return self.obs

    def select_query(self) -> None:
        # check valid query
        assert self.query is not None, "No query set!"
        # add to samples
        self.samples.append(self.query)
        self.total_samples += 1
        # avoid selecting the query again
        self.query = None

    @property
    def query_index(self) -> int:
        return len(self.samples)

    @property
    def query_size_reached(self) -> bool:
        return self.query_index >= self.query_size

    @property
    def budget_exhausted(self) -> bool:
        return self.total_samples >= self.budget


class StreamBasedEnv(gym.Env):
    """ Stream-based Active Learning Environment for Natural Language Processing Tasks
        using Transformer Models and Policies.

        Stream-based refers to the setup that at each point the policy is presented
        with one sample only for which it has to decide whether to include it in the
        next active learning step or not.
    """

    def __init__(
        self,
        # active learning hyperparameters
        budget:int,
        query_size:int,
        # active learning engine
        engine:ActiveLearningEngine,
        # metric used to compute reward
        metric:Metric,
        # query selection strategy
        query_strategy:AbstractStrategy,
        # data
        policy_pool_data:Dataset,
        model_pool_data:Dataset,
        model_test_data:Dataset,
        # data setup, inferred from data
        # and model when not specified
        policy_sequence_length:Optional[int] =None,
        model_sequence_length:Optional[int] =None,
        max_num_labels:Optional[int] =None
    ) -> None:
        # initialize environment
        super(StreamBasedEnv, self).__init__()
        # check data pools
        assert len(policy_pool_data) == len(model_pool_data), "Size mismatch between data pools!"

        # create state instance
        self.state = State(
            budget=min(budget, len(policy_pool_data)),
            query_size=query_size
        )

        # save al engine and create placeholder variable
        # for evaluator which is instantiated in reset
        self.engine = engine
        self.evaluator:Evaluator = None
        # save reference to reward metric
        self.metric = metric
        # save reference to query selection strategy
        self.query_strategy = query_strategy

        # save references to datasets
        self.policy_pool_data = policy_pool_data
        self.model_pool_data = model_pool_data
        self.model_test_data = model_test_data

        # specify data layout, i.e. the dimensions of the observation space
        self._policy_max_seq_len = policy_sequence_length or self._policy_seq_len
        self._model_max_seq_len = model_sequence_length or self._model_seq_len
        self._model_max_num_labels = max_num_labels or self._model_num_labels

    @property
    def _policy_seq_len(self) -> int:
        return len(self.policy_pool_data[0]['input_ids'])

    @property
    def _model_seq_len(self) -> int:
        return len(self.model_pool_data[0]['input_ids'])
    
    @property
    def _model_num_labels(self) -> int:
        return self.engine.trainer.unwrapped_model.config.num_labels

    @property
    def _vocab_size(self) -> int:
        # instead of actually checking the vocab size lets just say the
        # vocab size is the maximum number that can be represented by int64
        # use int64 since it matches the torch.long datatype
        return np.iinfo(np.int64).max
        
    @property
    def _test_data_loader(self) -> DataLoader:
        return DataLoader(
            self.model_test_data, 
            batch_size=self.engine.eval_batch_size,
            shuffle=False
        )

    @property
    def action_space(self) -> gym.spaces.Discrete:
        return gym.spaces.Discrete(2)

    @property
    def observation_space(self) -> gym.spaces.Dict:
        # build observation space
        # TODO: support variable number of labels to apply to same model to different datasets
        #       by introducing a logits mask or something similar
        return gym.spaces.Dict({
            # policy transformer inputs
            "input_ids": gym.spaces.Box(
                low=0, 
                high=self._vocab_size, 
                shape=(self._policy_max_seq_len,), 
                dtype=np.int64
            ),
            "attention_mask": gym.spaces.Box(
                low=0,
                high=1,
                shape=(self._policy_max_seq_len,),
                dtype=np.bool
            ),
            # model predictions
            "logits": gym.spaces.Box(
                low=-np.inf, 
                high=np.inf, 
                shape=(self._model_max_seq_len, self._model_max_num_labels),
                dtype=np.float32
            )
        })

    def _refill_queues(self) -> bool:
        # make sure queues are empty
        assert self.state.are_queues_empty, "Trying to refill non-empty queues!"
        
        # check if data pool is exhausted
        if len(self.loop.pool) == 0:
            return False

        # use initial sampling strategy as long as no active learning
        # step was done (here an activel learning step refers to training
        # the model)
        trained_once = self.engine.state.iteration > 0
        model_queue = self.loop.step() if trained_once else self.loop.init_step()

        # build policy queue corresponding to model queue    
        policy_queue = Subset(
            dataset=self.policy_pool_data, 
            indices=model_queue.indices
        )

        # get model predictions
        # note that by calling the evaluator's step function explicitly
        # no event are fired and thus the reward metric is not influenced
        batch = default_collate_drop_labels(model_queue)
        logits = self.evaluator.step(batch).logits.cpu()

        # combine policy queue and model predictions
        # and write queues to state
        self.state.model_queue = list(model_queue)
        self.state.obs_queue = [
            {
                'input_ids': policy_item['input_ids'].numpy(),
                'attention_mask': policy_item['attention_mask'].numpy(),
                'logits': logits[i, ...].numpy()
            } for i, policy_item in enumerate(policy_queue)
        ]
        
        return True

    def reset(self):
        """ Reset the environment and return initial observation """

        # reset internal state
        self.state.reset()

        # call engine started event, this resets the
        # internals of the engine, i.e. it re-initializes
        # the model and resets the optimizer, scheduler
        self.engine._fire_event(Events.STARTED)      
        self.engine._fire_event(Events.EPOCH_STARTED)
        # for now the state has to be reset manually
        # TODO: there has to be a way of doing this in event handler
        self.engine._reset_state()

        # initialize evaluator and attach metric if not done before
        if self.evaluator is None:
            self.evaluator = Evaluator(self.engine.trainer.unwrapped_model)
            self.metric.attach(self.evaluator, "__reward_metric")

        # initialize active loop which is used to gather queue
        # elements which one-by-one are passed to the policy
        self.loop = ActiveLoop(
            pool=self.model_pool_data,
            batch_size=self.engine.train_batch_size,
            # each queue is of size eval_batch_size since
            # the whole queue needs to be passed through
            # the model to gather the predictions
            query_size=self.engine.eval_batch_size,
            # queue elements are sampled randomly
            # TODO: it might make sense to use a better strategy
            #       in later stages of the rl run to make the
            #       decision for the policy more complex
            strategy=self.query_strategy,
            init_strategy=Random()
        )

        # set initial queues
        self._refill_queues()

        # return query observation
        return self.state.next_query()

    def step(self, action):
        
        if bool(action):
            # select current query
            self.state.select_query()
        
        # info dictionary
        info = {}
        # get next observations and check
        # if done, i.e. budget is exhausted
        obs = self.state.next_query()
        done = self.state.budget_exhausted

        # check if queues are empty
        if self.state.are_queues_empty and (not done):
            # refill if so
            done |= not self._refill_queues()

        # train on selected samples
        if self.state.query_size_reached or (done and (self.state.query_index > 0)):
            # need to update the iteration manually as only the step
            # function (i.e. the process function) is called
            self.engine.state.iteration += 1
            # also call the iteration started event
            self.engine._fire_event(Events.ITERATION_STARTED)
            
            # create dataset from aquired data
            samples = default_collate(self.state.samples)
            samples = NamedTensorDataset(**samples)
            # clear samples in state
            self.state.samples.clear()

            # do an active learning step with the newly aquired data
            self.engine.step(samples)

            # evaluate model on test data and get reward metric
            state = self.evaluator.run(self._test_data_loader)
            metric_val = state.metrics['__reward_metric']
            # compute reward and update state
            reward = metric_val - self.state.prev_metric
            self.state.prev_metric = metric_val
            
            # call iteration completed event
            # note that at this point the evaluator has ran
            # and metrics depending on the evaluator can be computed
            self.engine._fire_event(Events.ITERATION_COMPLETED)

        else:
            # delayed reward
            reward = 0
        
        if done:
            # call engine completed event
            self.engine._fire_event(Events.EPOCH_COMPLETED)
            self.engine._fire_event(Events.COMPLETED)
            # add engine metrics to info
            info['custom_metrics'] = self.engine.state.metrics

        # return observation, reward, done and info
        return obs, reward, done, info




import numpy as np
# import torch and ignite
import torch
import torch.nn.functional as F
from torch.utils.data import Subset
import ignite.distributed as idist
# import transformers
from transformers import (
    PreTrainedModel, 
    AutoTokenizer, 
    AutoModelForMaskedLM
)
# import sklearn
from sklearn.cluster import KMeans
from sklearn.neighbors import NearestNeighbors
# import base class and more
from .strategy import AbstractStrategy
from ..utils.data import move_to_device
from typing import Sequence, Any, Optional

class Alps(AbstractStrategy):
    """ Implementation of ALPS presented in `Cold-Start Active Learning through 
        Self-supervised Language Modeling` (Yuan et al., 2020).
    
        Args:
            model (PreTrainedModel): pretrained transformer model
            mlm_prob (Optional[float]): 
                masked-lanugage modeling probability used to generate 
                targets for loss that builds the surprisal embeddings
                Defaults to 15%.
            force_non_zero (Optional[bool]):
                wether to enforce non-zero surprisal embedding vectors.
                Surprisal embeddings are computed from the MLM pretraining,
                i.e. compute a loss for randomly sampled tokens. It might occur
                that no tokens are selected for a full instance (especially for
                short sequences), which leads to zero embeddings. Defaults to False.
    """

    def __init__(
        self,
        model:PreTrainedModel,
        mlm_prob:Optional[float] =0.15,
        force_non_zero:Optional[bool] =False
    ) -> None:
        # initialize strategy
        super(Alps, self).__init__()
        # check and save value
        assert 0 < mlm_prob <= 1, "Mask Probability must be in the interval (0, 1]"
        self.mlm_prob = mlm_prob
        self.force_non_zero = force_non_zero
        # get pretrained name
        pretrained_name_or_path = model.config.name_or_path
        assert pretrained_name_or_path is not None, "Expected pretrained model"
        # get special token ids
        tokenizer = AutoTokenizer.from_pretrained(pretrained_name_or_path)
        self.special_token_ids = torch.LongTensor(tokenizer.all_special_ids)
        # load the corresponding model with language model head and move to device(s)
        model = AutoModelForMaskedLM.from_pretrained(pretrained_name_or_path)
        self.model = idist.auto_model(model)

    @torch.no_grad()
    def process(self, batch:Any) -> torch.FloatTensor:
        """ Compute surpisal embeddings of given batch
            Args:
                batch (Any): input batch
            Returns:
                embed (torch.Tensor): surprisal embedding of samples in batch
        """
        # set model to evaluation mode
        self.model.eval()
        # move batch to device and get input ids
        batch = move_to_device(batch, device=idist.device())
        input_ids = batch['input_ids']

        # build label mask, i.e. find all candidate tokens for masking
        mask = input_ids.unsqueeze(-1) == self.special_token_ids.to(idist.device())
        mask = torch.any(mask, dim=-1)
        # check mask
        assert torch.any(mask, dim=-1).all(), "Invalid sequence of only special tokens found!"
        # compute probability of each token being masked
        mask_probs = (1.0 - mask.float()) * self.mlm_prob
        label_mask = torch.bernoulli(mask_probs).bool()
        non_zero_mask = torch.any(label_mask, dim=-1)
        # sample until all elements of the batch are valid,
        # i.e. there are masked elements in each
        while self.force_non_zero and (not non_zero_mask.all()):
            mask_probs[non_zero_mask, :] = 0
            label_mask |= torch.bernoulli(mask_probs).bool()
            non_zero_mask = torch.any(label_mask, dim=-1)

        # build labels from mask
        labels = input_ids.masked_fill(~label_mask, -100)

        # predict and compute mlm-loss
        # as described in the paper the input to the model is not masked
        # but the loss is only computed for a percentage of the tokens
        out = self.model(**batch)
        loss = F.cross_entropy(
            out.logits.flatten(end_dim=1), 
            labels.flatten(),
            reduction='none'
        ).reshape(labels.size())
        # make sure to avoid zero-embeddings
        if self.force_non_zero:
            assert torch.any(loss > 0.0, dim=-1).all(), "Found zero embedding vector!"
        # return normalized embeddings
        return F.normalize(loss, dim=-1)

    def sample(self, output:torch.FloatTensor, query_size:int) -> Sequence[int]:
        """ Select samples using kmeans clustering on surprisal embeddings 
        
            Args:
                output (torch.FloatTensor): surprisal embeddings
                query_size (int): number of samples to select

            Returns:
                indices (Sequence[int]): selected samples given by indices
        """
        assert output.ndim == 2
        # find cluster centers using kmeans
        kmeans = KMeans(n_clusters=query_size)
        kmeans = kmeans.fit(output.numpy())
        centers = kmeans.cluster_centers_
        # find unique nearst neighbors of cluster centers
        nearest = NearestNeighbors(n_neighbors=1)
        nearest = nearest.fit(output.numpy())
        centroids = nearest.kneighbors(centers, return_distance=False)
        centroids = set(centroids[:, 0])
        # handle doubles by filling up with random samples
        while len(centroids) < query_size:
            n = query_size - len(centroids)
            idx = np.random.randint(0, output.size(0), size=n)
            centroids |= set(idx)
        # return indices
        return centroids

class AlpsConstantEmbeddings(Alps):

    def query(
        self, 
        pool:Subset,
        query_size:int,
        batch_size:int
    ) -> Sequence[int]:
        """ Query samples to label using the strategy 
        
            Args:
                pool (Subset): pool of data points from which to sample
                query_size (int): number of data points to sample from pool
                batch_size (int): batch size used to process the pool dataset

            Returns:
                indices (Sequence[int]): indices of selected data points in pool
        """

        assert isinstance(pool, Subset), "Pool must be a subset to infer selected samples from dataset"
        # compute emebddings only once at the first call
        if self.output is None:
            return super(AlpsConstantEmbeddings, self).query(
                pool=pool,
                query_size=query_size,
                batch_size=batch_size
            )

        # remove items that were sampled in previous runs
        assert len(pool.dataset) == self.output.size(0), "Mismatch between embeddings and dataset"
        embeds = self.output[pool.indices]

        # sample from remaining embeddings
        return self.sample(embeds, query_size)



import gym
from torch.nn import Module
from abc import ABC, abstractmethod
from ray.rllib.utils.typing import ModelConfigDict

class FeatureExtractor(ABC, Module):
    """ Abstract Base Feature Extractor used by Models """

    def __init__(
        self,
        obs_space: gym.spaces.Space,
    ) -> None:
        # initialize module
        Module.__init__(self)

    @property
    @abstractmethod
    def feature_size(self) -> int:
        """ Dimension of the extracted features """
        raise NotImplementedError()



# torch
import torch
import torch.nn as nn
# import gym and rllib
import gym
from ray.rllib.models.modelv2 import ModelV2
from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
from ray.rllib.algorithms.dqn.dqn_torch_model import DQNTorchModel
from ray.rllib.utils.annotations import override
from ray.rllib.policy.rnn_sequencing import add_time_dimension
# import feature extractor
from ..extractors.extractor import FeatureExtractor
# typing
from typing import Type, List
from ray.rllib.utils.typing import TensorType, ModelConfigDict

class DQNModel(DQNTorchModel):
    
    @override(DQNTorchModel)
    def __init__(
        self,
        # observation and action space
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        # model setup and name
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        *,
        # feature extractor
        feature_extractor_type:Type[FeatureExtractor],
        feature_extractor_config:dict ={},
        # dqn model keyword arguments
        **kwargs
    ) -> None:

        # create feature extractor
        feature_extractor = feature_extractor_type(
            obs_space=obs_space,
            **feature_extractor_config
        )

        # get feature dimension
        feature_size = feature_extractor.feature_size
        
        # initialize super class
        DQNTorchModel.__init__(
            self,
            obs_space=obs_space,
            action_space=action_space,
            num_outputs=feature_size,
            model_config=model_config,
            name=name,
            **kwargs
        )

        # set feature extractor after module is initialized
        self.feature_extractor = feature_extractor
        
    @override(ModelV2)
    def forward(self, input_dict, state, seq_lens) -> (TensorType, List[TensorType]):
        # extract features
        return self.feature_extractor(input_dict), state
        
        
class ActorCriticModel(TorchModelV2, nn.Module):

    def __init__(
        self,
        # observation and action space
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        # model setup and name
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        *,
        # feature extractor
        feature_extractor_type:Type[FeatureExtractor],
        feature_extractor_config:dict ={},
        # prediction heads
        pi_hidden_layers:List[int] =[128],
        vf_hidden_layers:List[int] =[128],
    ) -> None:
        # initialize super classes
        TorchModelV2.__init__(self, obs_space, action_space, num_outputs, model_config, name)
        nn.Module.__init__(self)

        # create feature extractor
        self.feature_extractor = feature_extractor_type(
            obs_space=obs_space,
            **feature_extractor_config
        )

        # get feature dimension
        feature_size = self.feature_extractor.feature_size
        # action head
        self.pi = nn.Sequential(*[
            nn.Sequential(nn.Linear(n, m), nn.ReLU())
            for n, m in zip(
                [feature_size] + list(pi_hidden_layers),
                list(pi_hidden_layers) + [num_outputs]
            )
        ])
        # value-function head
        self.vf = nn.Sequential(*[
            nn.Sequential(nn.Linear(n, m), nn.ReLU())
            for n, m in zip(
                [feature_size] + list(vf_hidden_layers),
                list(pi_hidden_layers) + [1]
            )
        ])

        # set in forward call and used by value function
        self._features:TensorType = None

    @override(ModelV2)
    def forward(self, input_dict, state, seq_lens) -> (TensorType, List[TensorType]):
        # extract features and apply action head
        self._features = self.feature_extractor(input_dict)
        return self.pi(self._features), state

    @override(ModelV2)
    def value_function(self) -> TensorType:
        # check for features and apply value-function head
        assert self._features is not None, "No features found: make sure to call `forward` before `value_function`"
        return self.vf(self._features).squeeze(1)


class RecurrentActorCriticModel(ActorCriticModel):

    def __init__(
        self,
        # observation and action space
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        # model setup and name
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        # feature extractor
        feature_extractor_type:Type[FeatureExtractor],
        feature_extractor_config:dict ={},
        # prediction heads
        pi_hidden_layers:List[int] =[128],
        vf_hidden_layers:List[int] =[128],
        # rnn
        lstm_num_layers:int =1,
    ) -> None:
    
        # initialize actor critic module
        super(RecurrentActorCriticModel, self).__init__(
            # observation and action space
            obs_space=obs_space,
            action_space=action_space,
            # model setup and name
            num_outputs=num_outputs,
            model_config=model_config,
            name=name,
            # feature extractor
            feature_extractor_type=feature_extractor_type,
            feature_extractor_config=feature_extractor_config,
            # prediction heads
            pi_hidden_layers=pi_hidden_layers,
            vf_hidden_layers=vf_hidden_layers
        )

        # get feature dimension
        feature_size = self.feature_extractor.feature_size
        # create rnn
        self.lstm = nn.LSTM(
            input_size=feature_size,
            hidden_size=feature_size,
            num_layers=lstm_num_layers,
            bias=True,
            batch_first=True,
            dropout=0.0,
            # bi-directional doesn't make sense
            bidirectional=False
        )
        # create learnable initial lstm state
        self.h_0 = nn.Parameter(torch.zeros(lstm_num_layers, feature_size))
        self.c_0 = nn.Parameter(torch.zeros(lstm_num_layers, feature_size))

    @override(ModelV2)
    def get_initial_state(self):
        return [self.h_0, self.c_0]

    @override(ActorCriticModel)
    def forward(self, input_dict, state, seq_lens) -> (TensorType, List[TensorType]):

        # unpack state and transpose to match
        # expectation of lstm module
        h_i, c_i = state
        h_i = h_i.transpose(0, 1)
        c_i = c_i.transpose(0, 1)

        # extract features
        features = self.feature_extractor(input_dict)
        # get batch size
        n = features.size(0)

        # add time dimension        
        features = add_time_dimension(
            padded_inputs=features,
            max_seq_len=seq_lens.max(),
            framework='torch',
            time_major=False
        )

        # apply recurrent network
        features, (h_j, c_j) = self.lstm(features, (h_i, c_i))
        features = features.reshape(n, -1)

        # save features for value branch and compute logits
        self._features = features
        logits = self.pi(features)

        # reverse transposition to stay consistent
        h_j = h_j.transpose(0, 1)
        c_j = c_j.transpose(0, 1)

        # return policy output and new state
        return logits, [h_j, c_j]



# import torch
import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
# import ignite
from ignite.engine import Engine, Events
from ignite.handlers.stores import EpochOutputStore
# others
from ..utils.data import (
    map_tensors, 
    concat_tensors, 
    default_collate_drop_labels
)
from functools import partial
from abc import ABC, abstractmethod
from typing import Any, Sequence, Optional

class AbstractStrategy(Engine, ABC):
    """ Abstract base class for active learning strategies.
        Subclasses must define the following abstract methods:
            - `process`: executed by the ignite engine
            - `sample`: selects samples to label given the concatenated output of the engine run
    """

    def __init__(self):
        # initialize engine
        super(AbstractStrategy, self).__init__(type(self).process)
        # add output storage handler to self
        self.output_store = EpochOutputStore(
            # detach from gradient and move to cpu for storing
            output_transform=partial(map_tensors, fn=lambda t: t.detach().cpu())
        )
        self.output_store.attach(self)
        # last selected indices
        self._indices:Sequence[int] = None

    @property
    def output(self) -> Any:
        """ Output of the strategy engine. Returns output corresponding to the
            (unlabeled) data points of the pool last processed by the strategy.
            Depending on the type of strategy this corresponds to a tpye of
            representation (i.e. uncertainty scores, gradient embedding, etc.).
            Returns `None` before the first exection of the strategy.
        """
        # empty data store, i.e. strategy not executed yet
        if len(self.output_store.data) == 0:
            return None
        # get the recorded outputs and concatenate them
        return concat_tensors(self.output_store.data)

    @property
    def selected_indices(self) -> Sequence[int]:
        """ List of indices last selected by the strategy. Indices reference items
            in the unlabeled data pool last processed by the strategy. Returns
            `None` before the first execution. 
        """
        return self._indices

    @abstractmethod
    def process(self, batch:Any) -> Any:
        """ Process a given batch. Outputs of this function will 
            be concatenated and passed to the `sample` function. 

            Args:
                batch (Any): a batch of input samples from the unlabeled dataset

            Returns:
                output (Any): output to the batch
        """
        raise NotImplementedError()

    @abstractmethod
    def sample(self, output:Any, query_size:int) -> Sequence[int]:
        """ Sample examples given the concatenated output of the process function. 
            
            Args:
                output (Any): the concatenated output of the engine run, i.e. the process function.
                query_size (int): the number of samples to select for labeling

            Returns:
                indices (Sequence[int]): indices of samples in the unlabeled pool which to label next
        """
        raise NotImplementedError()

    def _sample(self, query_size:int, pool:Dataset) -> None:
        """ Event Handler called on `EPOCH_COMPLETED` (Note that there is no reason
            to run a strategy for more than one epoch). Calls the `sample` function
            and stores the output in the `_indices` attribute of the strategy.
            
            This way tracked time stored in `state.times[`COMPLETED`]` incorporates
            the processing and sampling.

            Only called when engine is run by calling `query` function. Not called
            when engine is directly executed by `run` function.
            
            Args:
                query_size (int): number of data points to sample from pool
                pool (Dataset): pool of data points from which to sample
        """
        # call sample function and check indices
        indices = list(self.sample(self.output, query_size))
        assert len(indices) == min(query_size, len(pool))
        # set attribute
        self._indices = indices
        
    def dataloader(self, data:Dataset, **kwargs) -> DataLoader:
        """ Create the dataloader for a given dataset with some specific configuration.            

            Args:
                data (Dataset): dataset to use
                **kwargs (Any): keyword arguments passed to the dataloader

            Returns:
                loader (DataLoader): dataloader from given dataset and configuration
        """
        return DataLoader(data, **kwargs)

    def query(
        self, 
        pool:Dataset,
        query_size:int,
        batch_size:int
    ) -> Sequence[int]:
        """ Query samples to label using the strategy 
        
            Args:
                pool (Dataset): pool of data points from which to sample
                query_size (int): number of data points to sample from pool
                batch_size (int): batch size used to process the pool dataset

            Returns:
                indices (Sequence[int]): indices of selected data points in pool
        """
        # check query size
        assert query_size <= len(pool), "Query size (%i) larger than pool (%i)" % (query_size, len(pool))
        # create dataloader
        loader = self.dataloader(
            pool,
            batch_size=batch_size,
            shuffle=False,
            collate_fn=default_collate_drop_labels
        )
        # add sample event handler which updates the `_indices` attribute
        handler = partial(self._sample, query_size=query_size, pool=pool)
        with self.add_event_handler(Events.EPOCH_COMPLETED, handler):
            # reset output store and run the engine
            self.output_store.reset()
            self.run(loader)
        # return the sampled indices
        return self._indices


class ScoreBasedStrategy(AbstractStrategy):
    """ Abstract Base Class for strategies that assign a score to each element
        in the pool. Implements the sample function of the abstract strategy.
        Subclasses must define the following abstract method: 
            - `process`: executed by the ignite engine

        Args:
            random_sample (Optional[bool]): 
                whether to random sample query according to distribution given by scores computed in `process` function.
                Defaults to False, meaning the elements with the maximal scores are selected.

    """

    def __init__(
        self,
        random_sample:Optional[bool] =False
    ) -> None:
        # initialize abstract strategy
        super(ScoreBasedStrategy, self).__init__()        
        # attach output check function
        self.add_event_handler(
            Events.ITERATION_COMPLETED, 
            type(self)._check_output
        )
        # whether to random sample according to
        # scores, alternatively use arg-max
        self.random_sample = random_sample

    def _check_output(self) -> None:
        """ Event handler checking the output. Makes sure the output is one-dimensional, 
            i.e. assignes a score to each element in the pool. 
        """
        # make sure output is a single score for each item in the current batch
        assert self.state.output.ndim == 1, "Expected scores to be one-dimensional but got shape %s" % str(tuple(self.state.output.size()))

    def sample(self, scores:torch.FloatTensor, query_size:int) -> Sequence[int]:
        """ Query samples to label using the strategy 
        
            Args:
                pool (Dataset): pool of data points from which to sample
                query_size (int): number of data points to sample from pool
                batch_size (int): batch size used to process the pool dataset

            Returns:
                indices (Sequence[int]): indices of selected data points in pool
        """
        # how many samples to draw
        k = min(scores.size(0), query_size)
        # draw samples from output scores
        return torch.multinomial(scores, k) if self.random_sample else scores.topk(k=k).indices



import torch
from ignite.metrics import Metric

class WorkSavedOverSampling(Metric):
    
    def reset(self):
        self.value = None

    def update(self, cm:torch.Tensor):
        # extract values from confusion matrix
        n = cm.sum()
        tp = torch.diag(cm)
        fn = cm.sum(dim=1) - tp
        fp = cm.sum(dim=0) - tp
        tn = n - tp - fp - fn
        # compute recall and wss score per class
        R = tp / (tp + fn + 1e-5)
        WSS = (tn + fn) / (n + 1e-5) - (1 - R)
        # average over classes and store
        self.value = WSS.mean()

    def compute(self) -> float:
        assert self.value is not None
        return self.value



import torch
import torch.nn as nn
import ignite.distributed as idist

from transformers import PreTrainedModel
from .strategy import ScoreBasedStrategy
from ..utils.data import move_to_device

from abc import abstractmethod
from typing import Sequence, Any, Optional

class UncertaintyStrategy(ScoreBasedStrategy):
    """ Abstract Base Class for uncertainty-based sampling strategies.
           
        Args:
            model (PreTrainedModel): transformer-based model used for prediction
            ignore_labels (Optional[Sequence[int]]): list of labels to ignore for computation of uncertainty scores
            random_sample (Optional[bool]): 
                whether to random sample query according to distribution given by uncertainty scores.
                Defaults to False, meaning the elements with maximum uncertainty scores are selected.
    """

    def __init__(
        self,
        model:PreTrainedModel,
        ignore_labels:Optional[Sequence[int]] =[],
        random_sample:Optional[bool] =False
    ) -> None:
        # initialize strategy
        super(UncertaintyStrategy, self).__init__(
            random_sample=random_sample
        )
        # move model to available device(s)
        self.model = idist.auto_model(model)
        # save labels to ignore as tensor
        self.ignore_labels = torch.LongTensor(ignore_labels)

    @abstractmethod
    def uncertainty_measure(self, probs:torch.FloatTensor) -> torch.FloatTensor:
        raise NotImplementedError()

    def reduce_scores(self, scores:torch.FloatTensor, mask:torch.BoolTensor) -> torch.FloatTensor:
        """ Reduce uncertainty scores for token classification tasks.
            By default uncertainty strategies in the context of token
            classification tasks compute one uncertainty score per token.
            The most naive way of reducing these into a single score is
            by computing their average, which is implemented in this function.

            Args:
                scores (torch.FloatTensor): 
                    uncertainty scores of shape (B, S, ...) where
                    B is the batch size and S is the sequence length.
                    Note that for nested nested token classification
                    (i.e. multiple classifications per token) the shape
                    is three-dimensional where the last dimension represents
                    the number of classifiers
                mask (torch.BoolTensor):
                    mask of valid scores in the `scores` tensor. The shape
                    matches the shape of the `score` tensor.

            Returns:
                reduced_scores (torch.FloatTensor): reduced scores of shape (B,)
        """
        # compute valid sequence lengths and avoid division by zero
        lengths = mask.sum(dim=1, keepdims=True)
        lengths = torch.maximum(lengths, torch.ones_like(lengths))
        # average over sequence and entities in case of nested bio tagging
        return (scores / lengths).flatten(start_dim=1).sum(dim=1)

    @torch.no_grad()
    def process(self, batch:Any) -> torch.FloatTensor:
        # move batch to device
        batch = move_to_device(batch, device=idist.device())
        # apply model and get output logits
        self.model.eval()
        logits = self.model(**batch).logits
        probs = torch.softmax(logits, dim=-1)
        # compute uncertainty scores
        scores = self.uncertainty_measure(probs)
        assert scores.size() == logits.size()[:scores.ndim]        

        # ignore labels
        _, labels = logits.max(dim=-1)
        ignore_mask = torch.isin(labels, self.ignore_labels.to(labels.device))
        scores[ignore_mask] = 0.0

        if scores.ndim == 1:
            # no reduction needed
            return scores

        # prepare mask for broadcasting with scores and ignore mask
        # this basically ensures that the mask applies to the first
        # dimensions, contrary to usual broadcasting which aligns
        # axes in reverse order
        mask = batch['attention_mask'].bool()
        mask = mask.reshape(mask.shape + (1,) * (scores.ndim - mask.ndim))
        assert mask.ndim == scores.ndim == ignore_mask.ndim

        # apply mask to scores
        scores.masked_fill_(~mask, 0.0)
        # reduce scores
        return self.reduce_scores(scores, mask & ~ignore_mask)

class LeastConfidence(UncertaintyStrategy):

    def uncertainty_measure(self, probs:torch.FloatTensor) -> torch.FloatTensor:
        return 1.0 - probs.max(dim=-1).values

class PredictionEntropy(UncertaintyStrategy):

    def uncertainty_measure(self, probs:torch.FloatTensor) -> torch.FloatTensor:
        return torch.special.entr(probs).sum(dim=-1)

class EntropyOverMax(UncertaintyStrategy):
    """ Uncertainty Sampling Strategy for token classification tasks. Computes the entropy over the maximum 
        predicted probability of each token. See Equation (4.1) in Joey Oehman (2021).

        Args:
            model (PreTrainedModel): transformer-based model used for prediction
            ignore_labels (Optional[Sequence[int]]): list of labels to ignore for computation of uncertainty scores

        References:
            - Joey Oehman, Active Learning for Named Entity Recognition with Swedish Language Models, 2021
    """    

    def uncertainty_measure(self, probs:torch.FloatTensor) -> torch.FloatTensor:
        # only makes sense for (nested) token-classification tasks
        assert probs.ndim >= 3, "Entropy over Maxima Sampling only makes sense for token classification tasks!"
        # only take the maximum over label space here to match expected shapes
        # in process and reduce score functions
        return probs.max(dim=-1).values
    
    def reduce_scores(self, scores:torch.FloatTensor, mask:torch.BoolTensor) -> torch.FloatTensor:
        # take maximum over classifiers/entities in case of nested
        # token classification tasks, for other tasks this does nothing
        scores = scores.reshape(scores.size(0), scores.size(1), -1).max(dim=-1).values
        # note that scores at ~mask are already set to zero
        # and entropy is defined to be zero at x=0 
        # (see https://pytorch.org/docs/stable/special.html)
        return torch.special.entr(scores).sum(dim=-1)



from ignite.metrics import Metric
from scipy.integrate import trapz

class AreaUnderLearningCurve(Metric):
    """ Compute area under the normalized learning curve.    
        Note that the x-Axis is normalized to range from 0
        to 1. Also the initial starting point (0, 0) is added
        on reset.
    """

    def reset(self) -> None:
        # save points on curve
        self.x = [0]
        self.y = [0]

    def update(self, output):
        # add point to lists
        self.x.append(output[0])
        self.y.append(output[1])

    def compute(self) -> float:
        # normalize x values (i.e. steps)
        # to range from 0 to 1
        x_max = max(self.x)
        x_norm = [x_ / x_max for x_ in self.x]
        # integrate over curve
        return trapz(self.y, x_norm)



import torch
from torch import Tensor
from torch.utils.data import TensorDataset, default_collate
# typing
import collections
from typing import Union, Any, Sequence, Dict, Callable

class NamedTensorDataset(TensorDataset):
    """ Dataset similar to `TensorDataset` but returns takes named tensors
        as keyword arguments and yields dictionaries instead of tuples.
    """

    def __init__(self, **named_tensors) -> None:
        self.names, tensors = zip(*named_tensors.items())
        super(NamedTensorDataset, self).__init__(*tensors)
    def __getitem__(self, idx) -> dict:
        tensors = super(NamedTensorDataset, self).__getitem__(idx)
        return dict(zip(self.names, tensors))

def default_collate_drop_labels(batch:Any) -> Any:
    # collate batch
    collated_batch = default_collate(batch)
    # pop labels if present
    if isinstance(collated_batch, collections.abc.Mapping):
        batch_no_labels = {key: collated_batch[key] for key in collated_batch if key != 'labels'}
        try:
            collated_batch = type(collated_batch)(batch_no_labels)
        except TypeError:
            return batch_no_labels
    # return collated batch
    return collated_batch

def map_tensors(
    tensors:Union[Dict[Any, Tensor], Sequence[Tensor], Tensor], 
    fn:Callable[[Tensor], Any]
) -> Union[Dict[Any, Tensor], Sequence[Tensor], Tensor]:
    
    # trivial case
    if isinstance(tensors, Tensor):
        return fn(tensors)

    # handler sequences
    elif isinstance(tensors, collections.abc.Sequence):
        moved_list = [map_tensors(t, fn=fn) for t in tensors]
        # try to convert it to the same type
        try:
            return type(tensors)(moved_list)
        except TypeError:
            # might not support initialization from interator
            return moved_list
    
    # handle mappings
    elif isinstance(tensors, collections.abc.Mapping):
        moved_dict = {key: map_tensors(tensors[key], fn=fn) for key in tensors}
        # try to convert it to the same type
        try:
            return type(tensors)(moved_dict)
        except TypeError:
            # might not support initialization from dict
            return moved_dict
    
    # fallback
    return tensors

def move_to_device(
    tensors:Union[Dict[Any, Tensor], Sequence[Tensor], Tensor], 
    device:torch.device
) -> Union[Dict[Any, Tensor], Sequence[Tensor], Tensor]:
    # move all tensors to the given device
    return map_tensors(tensors, fn=lambda t: t.to(device))

def concat_tensors(batch):
    """ torch's `default_collate` function but altered to concatenate instead of stack """
    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum(x.numel() for x in batch)
            storage = elem.storage()._new_shared(numel, device=elem.device)
            out = elem.new(storage).resize_(len(batch), *list(elem.size()))
        return torch.cat(batch, 0, out=out)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError()

            return concat_tensors([torch.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(elem, int):
        return torch.tensor(batch)
    elif isinstance(elem, string_classes):
        return batch
    elif isinstance(elem, collections.abc.Mapping):
        try:
            return elem_type({key: concat_tensors([d[key] for d in batch]) for key in elem})
        except TypeError:
            # The mapping type may not support `__init__(iterable)`.
            return {key: concat_tensors([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(concate_tensors(samples) for samples in zip(*batch)))
    elif isinstance(elem, collections.abc.Sequence):
        # check to make sure that the elements in batch have consistent size
        it = iter(batch)
        elem_size = len(next(it))
        if not all(len(elem) == elem_size for elem in it):
            raise RuntimeError('each element in list of batch should be of equal size')
        transposed = list(zip(*batch))  # It may be accessed twice, so we use a list.

        try:
            return elem_type([concat_tensors(samples) for samples in transposed])
        except TypeError:
            # The sequence type may not support `__init__(iterable)` (e.g., `range`).
            return [concat_tensors(samples) for samples in transposed]

    raise TypeError()



import torch
from transformers import PreTrainedTokenizer
from typing import List

class SequenceClassificationProcessor(object):

    def __init__(
        self, 
        tokenizer:PreTrainedTokenizer,
        max_length:int,
        text_column:str,
        label_column:str,
        **kwargs
    ) -> None:
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.text_column = text_column
        self.label_column = label_column

    def __call__(self, example):
        return self.tokenizer(
            text=example[self.text_column],
            max_length=self.max_length,
            truncation=True,
            padding='max_length',
            return_token_type_ids=False,
        ) | {'labels': example[self.label_column]}


class BioTaggingProcessor(object):

    def __init__(
        self,
        tokenizer:PreTrainedTokenizer,
        max_length:int,
        text_column:str,
        label_column:str,
        label_space:List[str],
        begin_tag_prefix:str,
        in_tag_prefix:int
    ) -> None:
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.text_column = text_column
        self.label_column = label_column
       
        # map label to id
        label2id = dict(zip(label_space, range(len(label_space))))
        # map begin/in tags to corresponding entities
        begin_tags = {tag: tag.lstrip(begin_tag_prefix) for tag in label_space if tag.startswith(begin_tag_prefix)}
        in_tags = {tag: tag.lstrip(in_tag_prefix) for tag in label_space if tag.startswith(in_tag_prefix)}
        # make sure there is a in-tag for each begin-tag and vise versa
        assert set(begin_tags.values()) == set(in_tags.values())
        # make sure corresponding begin- and in-tags reference the same entity
        assert all(entity == in_tags[tag] for tag, entity in in_tags.items())

        # map begin- to corresponding in-tag
        begin2in = {tag: in_tag_prefix + entity for tag, entity in begin_tags.items()}
        begin2in = [begin2in.get(tag, tag) for tag in label_space]
        begin2in = [label2id[tag] for tag in begin2in]
        # convert to tensor
        # this tensor maps the label-id of begin-tags to the label-id of the
        # corresponding in-tags. Label-ids of non-begin-tags remain untouched.
        # Examples:
        #    - begin2in[label2id["B-ORG"]] = label2id["I-ORG"]
        #    - begin2in[label2id["I-ORG"]] = label2id["I-ORG"]
        self.begin2in = torch.LongTensor(begin2in)

        # check tokenizer is fast
        if not tokenizer.is_fast:
            raise ValueError("Bio tagging processor needs fast pretrained tokenizer but got %s" % type(tokenizer).__name__)

    def __call__(self, example):
        
        # tokenize
        enc = self.tokenizer(
            text=example[self.text_column],
            add_special_tokens=True,
            padding='max_length',
            truncation=True,
            is_split_into_words=True,
            max_length=self.max_length,
            return_token_type_ids=False,
            return_attention_mask=True,
            return_overflowing_tokens=False,
            return_special_tokens_mask=True
        )

        # extract special tokens mask from encoding
        special_tokens_mask = torch.BoolTensor(enc.special_tokens_mask)

        # create word-ids tensor
        word_ids = torch.LongTensor([i for i in enc.word_ids() if i is not None])
        # create label-ids tensor and apply word-ids
        label_ids = torch.LongTensor(example[self.label_column])
        label_ids = label_ids[word_ids]

        # find tokens marked that are begin-tag but split into
        # word-pieces by tokenizer, only the first word-piece
        # should be annotated with the begin-tag however
        # the above assigns each wordpiece of the token
        # the same tag
        in_mask = (word_ids[:-1] == word_ids[1:])
        # note: this actually marks all wordpieces of
        # the same token except the first one no matter
        # the tag, however only the begin tags are altered
        # by the mapping below
        label_ids[1:][in_mask] = self.begin2in[label_ids[1:][in_mask]]

        # create the bio-tags which bascially are the
        # label-ids above with padding for special tokens
        bio_tags = torch.where(special_tokens_mask, -100, 0)
        bio_tags[~special_tokens_mask] = label_ids
 
        return {
            'input_ids': torch.LongTensor(enc.input_ids),
            'attention_mask': torch.LongTensor(enc.attention_mask),
            'labels': bio_tags
        }

class NestedBioTaggingProcessor(object):
    def __init__(
        self, 
        tokenizer:PreTrainedTokenizer, 
        max_length:int, 
        text_column:str, 
        label_column:str, 
        label_space:List[str],
        **kwargs
    ) -> None:
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.text_column = text_column
        self.label_column = label_column
        self.entity_types = label_space
        # 
        self.begin2in = torch.LongTensor([0, 2, 2])

    def __call__(self, example):
        
        # tokenize
        enc = self.tokenizer(
            text=example[self.text_column],
            add_special_tokens=True,
            padding='max_length',
            truncation=True,
            is_split_into_words=True,
            max_length=self.max_length,
            return_token_type_ids=False,
            return_attention_mask=True,
            return_overflowing_tokens=False,
            return_special_tokens_mask=True
        )
        
        # extract special tokens mask from encoding
        special_tokens_mask = torch.BoolTensor(enc.special_tokens_mask)

        # create word-ids tensor
        word_ids = torch.LongTensor([i for i in enc.word_ids() if i is not None])

        # build label ids tensor
        label_ids = torch.LongTensor([example[self.label_column][et] for et in self.entity_types])
        label_ids = label_ids[:, word_ids]
       
        # handle multiple word-pieces annotated as begin 
        in_mask = (word_ids[:-1] == word_ids[1:])
        label_ids[:, 1:][:, in_mask] = self.begin2in[label_ids[:, 1:][:, in_mask]]

        # create bio tags
        bio_tags = torch.where(special_tokens_mask, -100, 0)
        bio_tags = bio_tags.unsqueeze(1).repeat(1, label_ids.size(0))
        bio_tags[~special_tokens_mask, :] = label_ids.t()

        return {
            'input_ids': torch.LongTensor(enc.input_ids),
            'attention_mask': torch.LongTensor(enc.attention_mask),
            'labels': bio_tags
        }



import io
# import torch
import torch
import torch.nn as nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler, LambdaLR
from torch.utils.data import DataLoader
# import ignite
import ignite.distributed as idist
from ignite.engine import Engine, State, Events
from ignite.metrics import Average, Accuracy
from ignite.handlers import EarlyStopping, Checkpoint
from ignite.handlers.checkpoint import BaseSaveHandler
# transformers
from transformers import PreTrainedModel
# others
from ..core.utils.model import get_encoder_from_model
from ..core.utils.data import map_tensors, move_to_device
from .evaluator import Evaluator
from .schedulers import _TrainingStepsDependentMixin
from copy import deepcopy
from collections import OrderedDict
from typing import Optional, Iterable

class MemorySaveHandler(BaseSaveHandler, OrderedDict):
    
    def __init__(self, max_length:int =1) -> None:
        super(MemorySaveHandler, self).__init__()
        self.max_length = max_length

    def __call__(self, checkpoint, filename, metadata=None) -> None:
        # store checkpoint in bytes buffer instead of copying
        # the checkpoint to avoid copies on gpu memory
        buf = io.BytesIO()
        torch.save(checkpoint, buf)
        # save copy of checkpoint in memory dict
        self[filename] = buf
        # remove very last element if memory exceeds size
        if len(self) > self.max_length:
            self.popitem(last=False)

    def remove(self, filename):
        # remove object from memory
        self.pop(filename)

    def load(self, filename):
        # get buffer
        buf = dict.__getitem__(self, filename)
        buf.seek(0)
        # load buffer
        return torch.load(buf)


class Trainer(Evaluator):
    """ Engine to train a transformer model. Can be used for both active and passive learning loops.

        Args:
            model (PreTrainedModel): pre-trained transformer model to train
            optim (Optimizer): optimizer to use for parameter updates
            scheduler (Optional[_LRScheduler]): 
                learning rate scheduler. Defaults to LambdaLR(optim, lambda _: 1.0).
            acc_theshold (Optional[float]): 
                accuracy threshold, training is stopped if threshold is 
                surpassed on training data (Default: 1.0)
            patience (Optional[int]):
                early stopping patience, training is stopped if no improvement
                in accuracy on the validation data is achived for `patient` epochs.
                If set to `None`, no early stopping is applied (Default: None).
            incremental (Optional[bool]):
                whether to reuse the model trained in previous calls to `run` or
                to reset it to it's inital state. This also resets the optimizer
                and learning rate scheduler states. If set to false only the scheduler
                will be resetted. (Default: False)
            val_loader (Optional[DataLoader]):
                validation data loader. (Default: Train data loader)
    """

    def __init__(
        self, 
        model:PreTrainedModel, 
        optim:Optimizer, 
        scheduler:Optional[_LRScheduler] =None,
        acc_threshold:Optional[float] =1.0, 
        patience:Optional[int] =None,
        incremental:Optional[bool] =False,
        val_loader:DataLoader =None
    ) -> None:
        # save arguments
        self.acc_threshold = acc_threshold
        self.incremental = incremental
        
        # get model encoder
        self.encoder = get_encoder_from_model(model)
        # save optimizer and initialize evaluator
        self.optim = idist.auto_optim(optim)
        self.scheduler = scheduler or LambdaLR(optim, lambda _: 1.0)
        super(Trainer, self).__init__(model)
        # save validation dataloader
        self.val_loader = val_loader

        # save initial checkpoint
        self.init_encoder_ckpt = deepcopy(self.encoder.state_dict())
        self.init_optim_ckpt = deepcopy(self.optim.state_dict())
        self.init_scheduler_ckpt = deepcopy(self.scheduler.state_dict())
        # add event handler        
        self.add_event_handler(Events.STARTED, type(self)._load_init_ckpt)
        self.add_event_handler(Events.STARTED, type(self)._prepare_scheduler)

        # create train accuracy metric
        self.train_acc = Accuracy(output_transform=type(self).get_logits_and_labels)
        # and manually attach handlers to avoid writing accuracy in state
        self.add_event_handler(Events.EPOCH_STARTED, self.train_acc.started)
        self.add_event_handler(Events.ITERATION_COMPLETED, self.train_acc.iteration_completed)
        # add convergence event handler
        self.add_event_handler(Events.EPOCH_COMPLETED, type(self)._check_convergence)

        # create validation evaluator
        self.val_evaluator = Evaluator(model)
        Average(output_transform=lambda out: out['loss']).attach(self.val_evaluator, 'L')
        # validate after each epoch
        self.add_event_handler(Events.EPOCH_COMPLETED, type(self)._validate)

        # attach checkpoint handler saving best model parameters
        # w.r.t. validation loss
        self.ckpt = Checkpoint(
            to_save={
                'model': self.model,
                # also save evaluator to capture metrics corresponding to model
                'val-evaluator': self.val_evaluator
            },
            save_handler=MemorySaveHandler(),
            n_saved=1,
            score_function=lambda e: -e.state.metrics['L'],
        )
        self.val_evaluator.add_event_handler(Events.COMPLETED, self.ckpt)
        # add checkpoint event handlers
        self.add_event_handler(Events.STARTED, type(self)._reset_ckpt)
        self.add_event_handler(Events.COMPLETED, type(self)._load_best_ckpt)

        # attach to evaluator
        if patience is not None:
            # early stopping
            self.stopper = EarlyStopping(
                patience=patience,
                score_function=lambda e: -e.state.metrics['L'],
                trainer=self
            )
            # add event handlers
            self.add_event_handler(Events.STARTED, type(self)._reset_stopper)
            self.val_evaluator.add_event_handler(Events.COMPLETED, self.stopper)

    @property
    def converged(self) -> bool:
        """ Whether the training process converged """
        # convergence is met if engine terminated before maximum number of epochs
        return self.state.epoch < self.state.max_epochs

    @property
    def train_accuracy(self) -> float:
        """ Final accuracy on training dataset """
        return self.train_acc.compute()
    
    def step(self, batch):
        """ Training step function executed by the engine. """
        # move to device
        batch = move_to_device(batch, device=idist.device())
        # forward through model and compute loss
        self.model.train()
        self.optim.zero_grad()
        out = self.model.forward(**batch)
        # compute average loss
        if out['loss'].ndim >= 0:
            out['loss'] = out['loss'].mean()
        # optimizer parameters
        out['loss'].backward()
        self.optim.step()
        self.scheduler.step()
        # pass labels through
        out['labels'] = batch['labels']
        # detach and move to cpu
        return map_tensors(out, fn=lambda t: t.detach().cpu())

    def run(
        self, 
        data: Optional[Iterable] =None,
        max_epochs: Optional[int] =None,
        epoch_length: Optional[int] =None,
        min_epoch_length: Optional[int] =None
    ) -> State:
        """ Runs the trainer over the passed data.

            Overwrites the typical `ignite.Engine` behavior of continuing a
            previous run by resetting the state before starting the engine run.

            Args:
                data (Optional[Iterable]):
                    data (usually a `DataLoader`) to train the model on
                max_epochs (Optional[int]): 
                    maximum number of training epochs
                epoch_length (Optional[int]): 
                    length of a single training epoch. If provided, the epoch length
                    induced by the length of the dataloader is overwritten.
                min_epoch_length(Optional[int]): 
                    minimum length of a single training epoch. Overwrites the induced
                    epoch length only if it is too small. This argument is especially
                    useful in active learning scenarios where the first few training 
                    steps have only limited data and thus less update steps. Setting
                    `min_epoch_length` ensures a constant minumum number of updates.
            Returns:
                state (State): engine state after the run
        """
        # usually trainer tries to resume run if state holds unfinished runs
        # due to early stopping or other convergence criteria
        # to avoid that the state's progress is reset at the start of the run
        self.state = State()
        # get epoch length to use
        if min_epoch_length is not None:
            epoch_length = epoch_length or len(data)
            epoch_length = epoch_length if epoch_length >= min_epoch_length else min_epoch_length
        # run trainer
        return super(Trainer, self).run(
            data=data,
            max_epochs=max_epochs,
            epoch_length=epoch_length
        )

    def _load_init_ckpt(self, force:bool =False):
        """ Event handler to load the initial checkpoints. Called on `STARTED`. 

            Args:
                force (Optional[bool]): 
                    whether to force reset the model and optimizer even if
                    the incremental attribute is set to True.
        """
        if (not self.incremental) or force:
            # reset model by first re-initializing all weights
            # and loading the pretrained encoder afterwards
            self.unwrapped_model.init_weights()
            self.encoder.load_state_dict(self.init_encoder_ckpt)
            # reset optimizer state
            self.optim.load_state_dict(self.init_optim_ckpt)
        # always reset scheduler state
        self.scheduler.load_state_dict(self.init_scheduler_ckpt)
        # just making sure ;)
        self.optim.zero_grad(set_to_none=True)

    def _prepare_scheduler(self):
        """ Event handler to prepare the scheduler. Called on `STARTED`. """
        # update scheduler train steps
        if isinstance(self.scheduler, _TrainingStepsDependentMixin):
            self.scheduler.set_num_training_steps(
                steps=len(self.state.dataloader) * self.state.max_epochs
            )

    def _check_convergence(self):
        """ Event handler stopping training when training accuracy 
            threshold is surpassed. Called on 'EPOCH_COMPLTED'. 
        """
        # evaluate model on train set and check accuracy
        if (self.train_accuracy >= self.acc_threshold):
            self.terminate()

    def _validate(self):
        """ Event handler validating the model on the validation data loader.
            Falls back to the training dataloader if validation loader is not set.
            Called on 'EPOCH_COMPLETED'.
        """
        # use validation loader if given and fallback to train data loader
        self.val_evaluator.run(self.val_loader or self.state.dataloader)

    def _load_best_ckpt(self):
        """ Event handler loading the best checkpoint after training finished.
            Called on 'COMPLETED'.
        """
        # check if there is a checkpoint to load
        if self.ckpt.last_checkpoint is not None:
            # load checkpoint
            self.ckpt.load_objects(
                to_load={
                    'model': self.model,
                    'val-evaluator': self.val_evaluator
                },
                checkpoint=self.ckpt.save_handler.load(self.ckpt.last_checkpoint)
            )

    def _reset_ckpt(self):
        """ Event handler to reset the checkpoint. Called on `STARTED`. """
        self.ckpt.reset()

    def _reset_stopper(self):
        """ Event handler to reset the early stopper state. Callend on `STARTED`. """
        self.stopper.best_score = None
        self.stopper.counter = 0



from ray.rllib.algorithms.callbacks import DefaultCallbacks

class CustomMetricsFromEnvCallbacks(DefaultCallbacks):
    """ Callbacks to forward custom metrics from Environemnt
        info dict to episode
    """

    def on_episode_step(self, episode, **kwargs):
        # add custom metrics from environment to episode
        episode.custom_metrics.update(
            episode.last_info_for().get("custom_metrics", {})
        )

class LoggingCallbacks(DefaultCallbacks):
    """ Logging callbacks """

    def on_episode_start(self, episode, env_index, **kwargs):
        print("Episode {} (env-idx={}) started".format(
            episode.episode_id, env_index
        ))

    def on_episode_end(self, episode, env_index, **kwargs):
        print("Episode {} (env-idx={}) ended after {} steps".format(
            episode.episode_id, env_index, episode.length
        ))
        print("Custom Metrics: {}".format(episode.custom_metrics))

    def on_sample_end(self, samples, **kwargs):
        if samples.count > 0:
            print("Returned sample batch of size {}".format(samples.count))

    def on_train_result(self, algorithm, result, **kwargs):
        print("Algorithm.train() result: {} -> {} episodes".format(
            algorithm, result['episodes_this_iter']
        ))

    def on_learn_on_batch(self, policy, **kwargs):
        print("policy.learn_on_batch() result: {}".format(policy))



import gym
import torch
import transformers

from .extractor import FeatureExtractor

class TransformerFeatureExtractor(FeatureExtractor):

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        pretrained_ckpt:str,
    ) -> None:
        # initialize feature extractor
        super(TransformerFeatureExtractor, self).__init__(obs_space)
        # load pretrained transformer encoder
        self.encoder = transformers.AutoModel.from_pretrained(pretrained_ckpt)

    @property
    def feature_size(self) -> int:
        return self.encoder.config.hidden_size

    def forward(self, input_dict) -> torch.Tensor:
        # get observations from input
        input_ids = input_dict['obs']['input_ids'].long()
        attention_mask = input_dict['obs']['attention_mask'].long()

        # avoid invalid attention mask of all zeros, this seems to
        # happend when the environment is checked before actual
        # execution, also this should make sense as the first token
        # should always be the initial [CLS] special token
        attention_mask[:, 0] = 1

        # pass through transformer
        out = self.encoder(
            input_ids=input_ids,
            attention_mask=attention_mask
        )

        # return hidden state of [CLS]
        return out.last_hidden_state[:, 0, :]



import re
import torch
from transformers import PreTrainedModel
from collections import defaultdict
from typing import Optional, Dict

class TransformerParameterGroups(list):
    """ Helper class for Parameter Groups of transformer-based models. Used to initialize
        optimizers for transformer models. Adds basic hyperparameter-features like:
            - apply weight decay only to non-bias parameters
            - layer-wise learning rate decay

        Args:
            model (PreTrainedModel): pre-trained model to build parameter groups for
            lr (float): base learning rate
            lr_decay (Optional[float]): learning rate decay. Defaults to 1.0, i.e. constant learning rate.
            weight_decay (Optional[float]): weight decay rate. Default to 0.01.
    """

    def __init__(
        self,
        model:PreTrainedModel,
        lr:float,
        lr_decay:Optional[float] =1.0,
        weight_decay:Optional[float] =0.01
    ) -> None:
        # initialize list
        super(TransformerParameterGroups, self).__init__()

        # pattern to extract layer number of a
        # parameter from it's name
        pattern = re.compile(r".layer.(?P<num>[0-9]*).")
        num_index = pattern.groupindex['num']
        # group parameters into layers
        groups = defaultdict(dict)
        for name, param in model.named_parameters():
            match = pattern.search(name)
            layer = 'default' if match is None else int(match.group(num_index))
            groups[layer][name] = param

        # build parameter groups with set hyperparameters
        # start with default group
        self.add_parameters(
            named_parameters=groups.pop('default'),
            lr=lr,
            weight_decay=weight_decay
        )

        # get the total number of layers found
        # actually not the total number of but the biggest index
        num_layers = max(groups.keys())
        # add all layer parameters
        for i, params in groups.items():
            self.add_parameters(
                named_parameters=params,
                lr=lr * lr_decay**(num_layers - i),
                weight_decay=weight_decay
            )

    def add_parameters(
        self,
        named_parameters:Dict[str, torch.Tensor],
        lr:float,
        weight_decay:float
    ) -> None:
        """ Add parameter group for the given named parameters. Actually splits
            the parameters into bias and non-bias parameters and only applies
            weight decay to non-bias parameters.
        
            Args:
                named_parameters (Dict[str, torch.Tensor]): 
                    parameters to add to the group
                lr (float): 
                    learning rate for the parameter group
                weight_decay (float): 
                    weight decay for the parmaeter group. Only applied to
                    non-bias parameters
        """
        # non-biases with weight decay
        self.append(dict(
            params=[param for name, param in named_parameters.items() if not name.endswith('.bias')],
            lr=lr, weight_decay=weight_decay
        ))
        # disable weight decay for bias parameters
        self.append(dict(
            params=[param for name, param in named_parameters.items() if name.endswith('.bias')],
            lr=lr, weight_decay=0.0
        ))




import torch
import torch.nn as nn
from typing import Tuple, Any

class GoodfellowGradientNormForLinear(object):
    """ Goodfellow Gradient Norm for single linear layer.
        Attaches hooks to the given linear module. Gradient
        Norm is ready for computation after backward call.

        Gradient norms are computed for a single batch only
        and resetted for each forward call to the linear module.
        However multiple backward calls are supported.
    
        Can be used as a context manager.

        Args:
            module (nn.Linear): linear layer
    """
    
    def __init__(self, module:nn.Linear) -> None:
        assert isinstance(module, nn.Linear)
        # register capture hooks
        self.forward_handle = module.register_forward_hook(self.forward_hook)
        self.backward_handle = module.register_full_backward_hook(self.backward_hook)
        # reset
        self.H:torch.FloatTensor = None
        self.Z_bar:torch.FloatTensor = None

    def remove_hooks(self) -> None:
        """ Remove all hooks from module """
        # remove forward and backward hook
        self.forward_handle.remove()
        self.backward_handle.remove()

    @torch.no_grad()
    def forward_hook(
        self,
        module:nn.Linear,
        x:Tuple[torch.FloatTensor], 
        y:Tuple[torch.FloatTensor]
    ) -> None:
        """ Forward hook capturing inputs to module 
        
            Args:
                x (Tuple[torch.FloatTensor]): input to linear module
                y (Tuple[torch.FloatTensor]): output of linear module
        """
        # compute squared norm of input
        self.H = (x[0] * x[0]).flatten(start_dim=1).sum(dim=1)
        self.Z_bar = []

    @torch.no_grad()
    def backward_hook(
        self, 
        module:nn.Linear,
        x_grad:Tuple[torch.FloatTensor], 
        y_grad:Tuple[torch.FloatTensor]
    ) -> None:
        """ Backward hook caturing gradients of module outputs 
        
            Args:
                x_grad (Tuple[torch.FloatTensor]): gradient of input
                y_grad (Tuple[torch.FloatTensor]): gradient of output
        """
        # sum of squared gradients
        z_bar = (y_grad[0] * y_grad[0]).flatten(start_dim=1).sum(dim=1)
        self.Z_bar.append(z_bar)

    def compute(self) -> torch.FloatTensor:
        """ Compute Squarred Gradient Norm for last processed batch 
        
            Returns:
                norm (torch.FloatTensor): 
                    squared gradient norm for weight of linear module
                    in shape (b, m) where b is the batch size and m is
                    the number of backward calls.
        """
        # stack gradients from multiple backward calls
        # +1 to account for bias
        H = self.H.unsqueeze(1) + 1.0
        Z_bar = torch.stack(self.Z_bar, dim=1)
        # check dimensions
        assert H.size(0) == Z_bar.size(0), "Mismatch between forward (%i) and backward (%i) calls" % (H.size(0), Z_bar.size(0))
        return H * Z_bar

    def __enter__(self) -> "GoodfellowGradientNormForLinear":
        return self

    def __exit__(self, type:Any, value:Any, tb:Any) -> None:
        self.remove_hooks()


class GoodfellowGradientNorm(object):
    """ Compute Example-wise gradient norm of all linear components in the
        given module following Goodfellow (2016).
        
        Gradient norms are computed for a single batch only
        and resetted for each forward call to the linear module.
        However multiple backward calls are supported.
    
        Can be used as a context manager.
       
         Args:
            module (nn.Module): module which's linear components to analyze
    """
    
    def __init__(self, model:nn.Module) -> None:
        # register capturing hooks
        self.grad_captures = [
            GoodfellowGradientNormForLinear(module)
            for module in model.modules()
            if isinstance(module, nn.Linear)
        ]

    def remove_hooks(self) -> None:
        """ Remove all hooks from module """
        for grad in self.grad_captures:
            grad.remove_hooks()

    def compute(self) -> torch.FloatTensor:
        """ Compute Squarred Gradient Norm for last processed batch 
        
            Returns:
                norm (torch.FloatTensor): 
                    squared gradient norm for weight of linear module
                    in shape (b, m) where b is the batch size and m is
                    the number of backward calls.
        """
        # compute all gradients and add them
        return sum(grad.compute() for grad in self.grad_captures)
    
    def __enter__(self) -> "GoodfellowGradientNorm":
        return self

    def __exit__(self, type:Any, value:Any, tb:Any) -> None:
        self.remove_hooks()




__all__ = [
    "Random",
    "LeastConfidence",
    "PredictionEntropy",
    "EntropyOverMax",
    "Badge",
    "BadgeForSequenceClassification",
    "BadgeForTokenClassification",
    "Alps",
    "AlpsConstantEmbeddings",
    "EglByTopK",
    "EglBySampling"
]

from .strategy import (
    AbstractStrategy,
    ScoreBasedStrategy
)
from .random import Random
from .uncertainty import (
    LeastConfidence,
    PredictionEntropy,
    EntropyOverMax
)
from .badge import (
    Badge,
    BadgeForSequenceClassification,
    BadgeForTokenClassification
)
from .alps import (
    Alps,
    AlpsConstantEmbeddings
)
from .egl import (
    EglByTopK,
    EglBySampling
)



# import torch and ignite
import torch
import torch.nn as nn
import torch.nn.functional as F
import ignite.distributed as idist
# import transformers and kmeans++
from transformers import PreTrainedModel
from sklearn.cluster._kmeans import kmeans_plusplus
# import base and utils
from .strategy import AbstractStrategy
from ..utils.data import move_to_device
# import others
import warnings
from typing import Sequence, Tuple, Any

class Badge(AbstractStrategy):
    """ Implementation of `Deep Batch Active Learning By Diverse, 
        Uncertain Gradient Lower Bounds` (Ash et al., 2019) for 
        transformer models.

        Uses the final hidden state computed by the transformer model as 
        embedding, i.e. assumes a linear classifier on top of the hidden
        states. For Sequence Classification tasks the hidden state of the
        first token (usually [CLS]) is used. Note that this architecture is
        not always provided. As an example `BertForSequenceClassification`
        uses pooled outputs instead of the hidden states. To specify
        specific encoder see `BadgeForSequenceClassification` and 
        `BadgeForTokenClassification`.
        
        Pipeline:
        [input] -> encoder -> [out] -> [out.hidden_states[-1]], [out.logits]

        Args:
            model (PreTrainedModel): 
                pre-trained transformer-based classification model
            is_token_classification (bool):
                specifies the classification task, i.e. assumes
                token classification is set to `True` and sequence
                classification if set to `False`. By default tries
                to infer classification type by the model name.
    """

    def __init__(
        self,
        model:PreTrainedModel,
        is_token_classification:bool =None
    ) -> None:
        # initialize strategy
        super(Badge, self).__init__()
        # move model to available device(s)
        self.model = idist.auto_model(model)

        # save classification type
        self.is_tok_cls = is_token_classification
        # only if classification type is not given
        if self.is_tok_cls is None:
            # get the model task type
            name = type(model).__name__
            if name.endswith("ForSequenceClassification"):
                self.is_tok_cls = False
            elif name.endswith("ForTokenClassification"):
                self.is_tok_cls = True
            else:
                warnings.warn("Could not infer task type from model name. Assuming Sequence Classification Task!", UserWarning)
                self.is_tok_cls = False
   
        # test whether applying the classifier on the extracted embeddings
        # matches the logits predicted by the model
        if hasattr(model, 'classifier'):
            self._test_pipeline(model.classifier)

    def _test_pipeline(self, classifier:nn.Linear) -> None:
        # get model logits and embeddings
        target_logits, embeds = self._get_logits_and_embeds({
            'input_ids': torch.zeros((1, 1), dtype=int, device=idist.device())
        })
        # use embeddings to compute classifier logits
        classifier.eval()
        classifier = classifier.to(idist.device())
        classifier_logits = classifier(embeds)
        # compare
        if not torch.allclose(target_logits, classifier_logits):
            warnings.warn("Pipeline output doesn't match model output.", UserWarning)
        if not isinstance(classifier, nn.Linear):
            warnings.warn("Classifier must be linear.", UserWarning)

    def _get_logits_and_embeds(self, batch) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        """ Get the logits predicted by the model and the embeddings which 
            are used for classification.
            
            Args:
                batch (Any): input batch to apply the model to
                
            Returns:
                logits (torch.Tensor): predited logits
                emebds (torch.Tensor): embeddings used for classification
        """
        # apply model
        self.model.eval()
        out = self.model(**batch, output_hidden_states=True)
        # get embeds, use encoding of [CLS] token for sequence classification
        hidden_states = out.hidden_states[-1]
        embeds = hidden_states if self.is_tok_cls else hidden_states[:, 0, :]
        # return logits and embeddings
        return out.logits, embeds

    @torch.no_grad()
    def process(self, batch:Any) -> torch.FloatTensor:
        """ Compute the hallucinated gradient embeddings of a given batch
        
            Args:
                batch (Any): input batch
            Returns:
                g (torch.Tensor): hallucinated gradient embedding of samples in batch
        """
        # move batch to device and apply model
        batch = move_to_device(batch, device=idist.device())
        logits, embeds = self._get_logits_and_embeds(batch)
        # compute predictions
        probs = torch.softmax(logits, dim=-1)
        preds = torch.argmax(probs, dim=-1)
        # compute gradient embedding
        w = -(probs - F.one_hot(preds, num_classes=probs.size(-1)))
        if self.is_tok_cls and ('attention_mask' in batch):
            # apply attention mask in case of token classification
            mask = batch['attention_mask'].bool()
            w[~mask] = 0.0
        g = w.unsqueeze(-1) * embeds.unsqueeze(-2)
        # return gradient embedding
        return g.flatten(start_dim=1)
    
    def sample(self, output:torch.FloatTensor, query_size:int) -> Sequence[int]:
        """ Select samples using kmeans++ on the hallucinated gradient embeddings.

            Args:
                output (torch.Tensor): hallucinated gradient emebddings
                query_size (int): number of samples to select

            Returns:
                indices (Sequence[int]): selected samples given by their indices
        """
        # use kmeans++ algorithm to find spread samples
        # w.r.t. the computed gradient embedding
        _, indices = kmeans_plusplus(
            X=output.numpy(),
            n_clusters=query_size
        )
        # return the indices of the selected samples
        return indices

class BadgeForSequenceClassification(Badge):
    """ Implementation of `Deep Batch Active Learning By Diverse, 
        Uncertain Gradient Lower Bounds` (Ash et al., 2019) for 
        transformer-based sequence classification models.

        Uses the `pooler_output` field of the encoder output as
        embeddings which are passed to the linear classifier. If
        the field is not specified the hidden state of the first
        token (usually [CLS]) is used as embedding.

        Pipeline:
        [input] -> encoder -> [out] -> [out.pooler_output] -> classifier -> [logits]

        Args:
            encoder (PreTrainedModel): transformer-based encoder
            classifier (nn.Linear): linear classifier
    """

    def __init__(
        self,
        encoder:PreTrainedModel,
        classifier:nn.Linear
    ) -> None:
        # initialize strategy
        AbstractStrategy.__init__(self)
        # move encoder to device(s)
        self.encoder = idist.auto_model(encoder)
        # make sure the classifier is linear and move
        # it to available device(s)
        assert isinstance(classifier, nn.Linear)
        self.classifier = idist.auto_model(classifier)
        # save classification type
        self.is_tok_cls = False

    def _get_logits_and_embeds(self, batch:Any) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        # set model to eval mode
        self.encoder.eval()
        self.classifier.eval()
        # apply encoder and extract embeddings from output
        # fallback to encoding of [CLS] token is pooler output is not specified
        out = self.encoder(**batch, output_hidden_states=True)
        embeds = out.hidden_states[-1][:, 0, :] if not hasattr(out, 'pooler_output') else \
            out.hidden_states[-1][:, 0, :] if out.pooler_output is None else \
            out.pooler_output
        # warn about fallback
        if not hasattr(out, 'pooler_output'):
            warnings.warn("Model output does not specify pooler output. Using final hidden state of [CLS] token instead.", UserWarning)
        # apply classifier and return
        return self.classifier(embeds), embeds

class BadgeForTokenClassification(Badge):
    """ Implementation of `Deep Batch Active Learning By Diverse, 
        Uncertain Gradient Lower Bounds` (Ash et al., 2019) for 
        transformer-based token classification models.

        Uses the final hidden states produced by the encoder as
        embeddings which are passed to the linear classifier. 

        Pipeline:
        [input] -> encoder -> [out] -> [out.pooler_output] -> classifier -> [logits]

        Args:
            encoder (PreTrainedModel): transformer-based encoder
            classifier (nn.Linear): linear classifier
    """
    
    def __init__(
        self,
        encoder:PreTrainedModel,
        classifier:nn.Linear
    ) -> None:
        # initialize strategy
        AbstractStrategy.__init__(self)
        # move encoder to device(s)
        self.encoder = idist.auto_model(encoder)
        # make sure the classifier is linear and move
        # it to available device(s)
        assert isinstance(classifier, nn.Linear)
        self.classifier = idist.auto_model(classifier)
        # save classification type
        self.is_tok_cls = True
    
    def _get_logits_and_embeds(self, batch:Any) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        # set model to eval mode
        self.encoder.eval()
        self.classifier.eval()
        # apply encoder and extract embeddings from output
        out = self.encoder(**batch, output_hidden_states=True)
        embeds = out.hidden_states[-1]
        # apply classifier and return
        return self.classifier(embeds), embeds




import warnings
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler, LambdaLR

class _TrainingStepsDependentMixin(_LRScheduler):

    @property
    def num_training_steps(self) -> int:
        if not hasattr(self, '_num_training_steps'):
            warnings.warn("Number of training steps not set!", UserWarning)
            return -1
        return self._num_training_steps

    @num_training_steps.setter
    def num_training_steps(self, steps:int) -> None:
        self.set_num_training_steps(steps)

    def set_num_training_steps(self, steps:int) -> None:
        self._num_training_steps = steps

class LinearWithWarmup(LambdaLR, _TrainingStepsDependentMixin):

    def __init__(
        self,
        optim:Optimizer,
        warmup_proportion:float,
        last_epoch:int =-1,
        verbose:bool =False
    ) -> None:
        # save warmup proportion
        # number of training steps is handled by mixin
        self.warmup_proportion = warmup_proportion
        # initialize scheduler
        super(LinearWithWarmup, self).__init__(
            optimizer=optim,
            lr_lambda=self._lr_lambda,
            last_epoch=last_epoch,
            verbose=verbose
        )

    @property
    def num_warmup_steps(self) -> int:
        return int(self.num_training_steps * self.warmup_proportion)

    def _lr_lambda(self, current_step:int) -> float:
        # get total number of training steps and warmup steps
        total_steps = self.num_training_steps
        warmup_steps = self.num_warmup_steps
        # compute learning rate scaling factor
        return (
            # warmup
            (float(current_step) / max(1.0, float(warmup_steps))) if current_step < warmup_steps else \
            # linear decay
            max(0.0, float(total_steps - current_step) / max(1.0, float(total_steps - warmup_steps)))
        )



from setuptools import setup, find_packages

setup(
    name="active-transformers",
    version="0.2.0",
    description="Active Learning for Transformer with focus on Sequence Tagging tasks",
    long_description=open("README.md", "r").read(),
    long_description_content_type='text/markdown',
    license="Apache 2.0",
    classifiers=[
        "Environment :: GPU :: NVIDIA CUDA",
        "Intended Audience :: Science/Research",
        "License :: Freely Distributable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.9",
    ],

    author="Niclas Doll",
    author_email="niclas@amazonis.net",
    url="https://github.com/ndoll1998/active-transformers/tree/master",

    packages=find_packages(exclude=["tests"]),
    package_dir={
        "active": "active",
    },

    python_requires=">=3.9",
    install_requires=[
        "torch>=1.11.0",
        "pytorch-ignite>=0.4.9",
        "transformers>=4.19.2",
        "datasets>=2.7.1",
        "scikit-learn>=1.1.1",
        "scipy>=1.9.0",
        "matplotlib>=3.5.2",
        "pydantic>=1.10.2",
        "ray[default]>=2.1.0"
    ],
    extras_require={
        "rl": ["ray[rllib]>=2.1.0"]
    },

    tests_require=["pytest>=7.1.2"],
    entry_points={
        "console_scripts": [
            "active.train = active.scripts.run_train:main",
            "active.active = active.scripts.run_active:main",
            "active.rl = active.scripts.run_rl_active:main"
        ]
    }
) 



import torch
import torch.nn as nn
# import ignite
import ignite.distributed as idist
from ignite.engine import Engine, Events
# transformers
from transformers import PreTrainedModel
# others
from ..core.utils.data import move_to_device

class Evaluator(Engine):
    """ Engine evaluating a transformer model on a given dataset.

        Args:
            model (PreTrainedModel): pre-trained transformer model to evaluate
    """

    def __init__(
        self,
        model:PreTrainedModel,
    ) -> None:
        # save model and initialize engine
        self.model = idist.auto_model(model)
        super(Evaluator, self).__init__(type(self).step)

    @property
    def unwrapped_model(self) -> PreTrainedModel:
        """ The pure transformer model """
        if isinstance(self.model, PreTrainedModel):
            return self.model
        elif isinstance(self.model, nn.DataParallel):
            return self.model.module
        raise ValueError("Unexpected model type: %s" % type(self.model))

    @torch.no_grad()
    def step(self, batch):
        """ Evaluation step function """
        # move to device
        batch = move_to_device(batch, device=idist.device())
        # forward through model
        self.model.eval()
        out = self.model.forward(**batch)
        # pass labels through for metrics
        if 'labels' in batch:
            out['labels'] = batch['labels']
        # compute average loss over dataparallel model output
        if ('loss' in out) and (out['loss'] is not None) and (out['loss'].ndim >= 0):
            out['loss'] = out['loss'].mean()
        # back to cpu to clear device memory
        return move_to_device(out, device=torch.device('cpu'))

    @staticmethod
    def get_logits_and_labels(out:dict):
        """ Function to extract the logits and labels from the model output.
            Can be used as `output_transform` in ignite metrics.
        
            Args:
                out (dict): 
                    output of a forward call to the `PreTrainedModel`.

            Returns:
                logits (torch.Tensor): logits tensor only containing valid entries
                labels (torch.Tensor): labels tensor only containing valid entries
        """
        logits, labels = out['logits'], out['labels']
        mask = (labels >= 0)
        return logits[mask, :], labels[mask]



import torch
import numpy as np
from torch.utils.data import DataLoader, random_split
# import trainer and evaluator engine
from active.helpers.trainer import Trainer
from active.helpers.evaluator import Evaluator
# import data processors
from active.helpers.processor import (
    BioTaggingProcessor,
    NestedBioTaggingProcessor,
    SequenceClassificationProcessor
)
# import ignite
from ignite.engine import Events
from ignite.contrib.handlers.tqdm_logger import ProgressBar
from ignite.metrics import Recall, Precision, Average, Fbeta, Accuracy
# hugginface
import datasets
import transformers
# others
import wandb
from enum import Enum
from typing import Tuple, Dict, Union, Callable, Type
from pydantic import BaseModel, validator, root_validator

def attach_metrics(engine, tag=None):
    # create metrics
    L = Average(output_transform=lambda out: out['loss'])
    A = Accuracy(output_transform=type(engine).get_logits_and_labels)
    R = Recall(output_transform=type(engine).get_logits_and_labels, average=True)
    P = Precision(output_transform=type(engine).get_logits_and_labels, average=True)
    F = Fbeta(beta=1.0, output_transform=type(engine).get_logits_and_labels, average=True)
    # attach metrics
    L.attach(engine, 'L' if tag is None else ('%s/L' % tag))
    A.attach(engine, 'A' if tag is None else ('%s/A' % tag))
    R.attach(engine, 'R' if tag is None else ('%s/R' % tag))
    P.attach(engine, 'P' if tag is None else ('%s/P' % tag))
    F.attach(engine, 'F' if tag is None else ('%s/F' % tag))

class ForwardForNestedTokenClassification(transformers.PreTrainedModel):
    """ Forward container class used by transformer models for nested token classification.
        Has to be combined with a transformer model for token classification.
    """

    def forward(self, *args, **kwargs):
        """ Forward function for nested token classification.
            Based on forward for standard token classification.
        
            Output logits are of shape (B, S, E, 3) where B
            is the batch size, S is the sequence length and
            E is the number of classifiers/entities to predict.
            The final dimension corresponds to BIO labeling scheme.    
        """
        # expects dictionary output
        kwargs['return_dict'] = True
        # pop labels from arguments if given
        # to avoid loss-computation in parent
        labels = kwargs.pop('labels') if 'labels' in kwargs else None
        
        # call forward transformer for token classification and reorganize logits
        out = super(ForwardForNestedTokenClassification, self).forward(*args, **kwargs)
        out['logits'] = out.logits.reshape(out.logits.size(0), out.logits.size(1), -1, 3)

        # compute loss
        if labels is not None:
            out['loss'] = torch.nn.functional.cross_entropy(out.logits.reshape(-1, 3), labels.reshape(-1))
        
        # return output
        return out 
        
class AutoModelForNestedTokenClassification(object):
    
    @staticmethod
    def from_pretrained(pretrained_ckpt, **kwargs):
        # re-interpret num labels as num entities
        # for each entity introduce 3 labels (i.e. B,I,O)
        kwargs['num_labels'] = 3 * kwargs.get('num_labels', 1)

        # get requested model type        
        config = transformers.AutoConfig.from_pretrained(pretrained_ckpt)
        model_type = transformers.AutoModelForTokenClassification._model_mapping.get(type(config), None)
        # create new model type overwriting the forward method
        # for nested bio tagging predictions and loss computation
        model_type_for_nested = type(
            "TransformerForNestedTokenClassification",
            (ForwardForNestedTokenClassification, model_type),
            {}
        )

        return model_type_for_nested.from_pretrained(pretrained_ckpt, **kwargs)

class Task(Enum):
    """ Enum defining supported task types:
            (1) sequence:    marks the experiment as a sequence classification task
            (2) bio-tagging: marks the experiment as a bio-tagging task
    """
    SEQUENCE = "sequence"
    BIO_TAGGING = "bio-tagging"
    NESTED_BIO_TAGGING = "nested-bio-tagging"

    @property
    def model_type(self) -> Type[transformers.PreTrainedModel]:
        if self is Task.SEQUENCE:
            return transformers.AutoModelForSequenceClassification
        elif (self is Task.BIO_TAGGING):
            return transformers.AutoModelForTokenClassification
        elif (self is Task.NESTED_BIO_TAGGING):
            return AutoModelForNestedTokenClassification

        raise ValueError("No model found for task <%s>" % self)
    
    @property
    def processor_type(self) -> Type[Callable]:
        if self is Task.SEQUENCE:
            return SequenceClassificationProcessor
        elif self is Task.BIO_TAGGING:
            return BioTaggingProcessor
        elif self is Task.NESTED_BIO_TAGGING:
            return NestedBioTaggingProcessor
 
        raise ValueError("No data processor found for task <%s>" % self)

class DataConfig(BaseModel):
    """ Data Configuration Model """
    task:Task =None
    # dataset config
    dataset:str
    text_column:str
    label_column:str
    # bio-tag prefixes
    begin_tag_prefix:str =None
    in_tag_prefix:str =None
    # sequence length
    min_sequence_length:int
    max_sequence_length:int

    @root_validator()
    def _check_config(cls, values):
        # get values
        task = values.get('task')
        dataset = values.get('dataset')
        text_col = values.get('text_column')        
        label_col = values.get('label_column')        
        min_seq_len = values.get('min_sequence_length')
        max_seq_len = values.get('max_sequence_length')
        begin_tag_prefix = values.get('begin_tag_prefix')
        in_tag_prefix = values.get('in_tag_prefix')

        # check if dataset is none
        if dataset is None:
            raise ValueError("Dataset set to None")

        try:
            # try to load dataset builder
            builder = datasets.load_dataset_builder(dataset)
        except FileNotFoundError as e:
            # handle invalid/unkown datasets
            raise ValueError("Unkown dataset: %s" % dataset) from e
        
        # get dataset info
        info = builder._info()

        # check if label column is valid
        if label_col not in info.features:
            raise ValueError("Specified label column `%s` is not present in dataset. Valid columns are: %s" % (label_col, ','.join(info.features.keys())))

        # check if text column is valid
        if text_col not in info.features:
            raise ValueError("Specified text column `%s` is not present in dataset. Valid columns are: %s" % (text_col, ','.join(info.features.keys())))

        # make sure sequence length range is set properly
        if min_seq_len > max_seq_len:
            raise ValueError("Minimum sequence length (%i) larger than maximum sequence length (%i)" % (min_seq_len, max_seq_len))

        # make sure the tag prefixes are set for bio tagging tasks
        if (task is not None) and (task is Task.BIO_TAGGING):
            if (begin_tag_prefix is None) or (in_tag_prefix is None):
                raise ValueError("Begin/In tag prefixes must be set for bio tagging tasks!")

        # all done
        return values

    #
    # Helper Functions
    #

    @property
    def dataset_info(self) -> datasets.DatasetInfo:
        builder = datasets.load_dataset_builder(self.dataset)
        return builder._info()

    @property
    def label_space(self) -> Tuple[str]:
        if self.task is Task.SEQUENCE:
            return tuple(self.dataset_info.features[self.label_column].names)
        elif self.task is Task.BIO_TAGGING:
            return tuple(self.dataset_info.features[self.label_column].feature.names)
        elif self.task is Task.NESTED_BIO_TAGGING:
            # for nested bio tagging tasks the label space
            # generated here is actually only the entity types
            # keep in mind that for each unique entity type
            # the Begin, In and Out-tags are introduces
            # the overall label space is three times larger
            # (see `AutoModelForNestedTokenClassification.from_pretrained`)
            return tuple(self.dataset_info.features[self.label_column].keys())

    def load_dataset(
        self, 
        tokenizer:transformers.PreTrainedTokenizer,
        *,
        split:Dict[str, str] ={'train': 'train', 'test': 'test'}, 
        use_cache:bool =False
    ) -> Dict[str, datasets.arrow_dataset.Dataset]:
        # load dataset 
        ds = datasets.load_dataset(self.dataset, split=split)
        
        # create processor
        processor = self.task.processor_type(
            tokenizer=tokenizer,
            max_length=self.max_sequence_length,
            text_column=self.text_column,
            label_column=self.label_column,
            label_space=self.label_space,
            begin_tag_prefix=self.begin_tag_prefix,
            in_tag_prefix=self.in_tag_prefix         
        )
        # create filter function
        pad_token_id = tokenizer.pad_token_id
        filter_ = lambda e: (np.asarray(e['input_ids']) != pad_token_id).sum() > self.min_sequence_length

        # preprocess datasets
        ds = {
            key: dataset \
                .map(processor, batched=False, desc=key, load_from_cache_file=use_cache) \
                .filter(filter_, batched=False, desc=key, load_from_cache_file=use_cache)
            for key, dataset in ds.items()
        }

        # set data formats
        for dataset in ds.values():
            dataset.set_format(
                type='torch',
                columns=['input_ids', 'attention_mask', 'labels']
            )
        
        # return preprocessed datasets
        return ds

class ModelConfig(BaseModel):
    """ Model Configuration Model """
    task:Task = None
    # pretrained model checkpoint
    pretrained_ckpt:str
    
    @validator('pretrained_ckpt')
    def _check_pretrained_ckpt(cls, value):
        try:
            # check if model is valid by loading config
            transformers.AutoConfig.from_pretrained(value)
        except OSError as e:
            # handle model invalid
            raise ValueError("Unkown pretrained checkpoint: %s" % value) from e
 
        return value

    @property
    def tokenizer(self) -> transformers.PreTrainedTokenizer:
        return transformers.AutoTokenizer.from_pretrained(self.pretrained_ckpt, use_fast=True, add_prefix_space=True)

    def load_model(self, label_space:Tuple[str]) -> transformers.PreTrainedModel:
        return self.task.model_type.from_pretrained(
            self.pretrained_ckpt, 
            num_labels=len(label_space)
        )

class TrainerConfig(BaseModel):
    """ Trainer Configuration Model """
    incremental:bool
    # optimizer setup
    lr:float
    weight_decay:float
    # training setup
    batch_size:int
    max_epochs:Union[int, None]
    epoch_length:Union[int, None]
    min_epoch_length:Union[int, None]
    # convergence criteria
    early_stopping_patience:int
    accuracy_threshold:float

    def build_trainer(self, model:transformers.PreTrainedModel):
        # create optimizer
        optim=torch.optim.AdamW(model.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        # create trainer
        return Trainer(
            model=model,
            optim=optim,
            scheduler=None,
            acc_threshold=self.accuracy_threshold,
            patience=self.early_stopping_patience,
            incremental=self.incremental
        )

    @property
    def run_kwargs(self) -> Dict[str, int]:
        return self.dict(include={'max_epochs', 'epoch_length', 'min_epoch_length'})

class ExperimentConfig(BaseModel):
    """ Experiment Configuration Model
        Includes all of the above configurations as sub-models
    """
    # experiment name used for referencing
    # and task type
    name:str
    task:Task
    # data, model and trainer config
    data:DataConfig
    model:ModelConfig
    trainer:TrainerConfig

    @validator('data', 'model', pre=True)
    def _pass_task_to_sub_configs(cls, v, values):
        assert 'task' in values
        if isinstance(v, BaseModel):
            return v.copy(update={'task': values.get('task')})
        elif isinstance(v, dict):
            return v | {'task': values.get('task')}

    def load_dataset(self, **kwargs) -> Dict[str, datasets.arrow_dataset.Dataset]:
        return self.data.load_dataset(self.model.tokenizer, **kwargs)

    def load_model(self) -> transformers.PreTrainedModel:
        return self.model.load_model(self.data.label_space)

    def build_trainer(self) -> Trainer:
        return self.trainer.build_trainer(self.load_model())

def main():
    from argparse import ArgumentParser
    # build argument parser
    parser = ArgumentParser(description="Train transformer model on sequence or token classification tasks.")
    parser.add_argument("--config", type=str, required=True, help="Path to a valid experiment configuration")
    parser.add_argument("--use-cache", action='store_true', help="Load cached preprocessed datasets if available")
    parser.add_argument("--seed", type=int, default=1337, help="Random Seed")
    # parse arguments
    args = parser.parse_args()

    # set random seed
    torch.manual_seed(args.seed)
    np.random.seed(args.seed)

    # parse configuration
    config = ExperimentConfig.parse_file(args.config)
    print("Config:", config.json(indent=2))

    # load datasets and build trainer
    ds = config.load_dataset(use_cache=args.use_cache)
    trainer = config.build_trainer()
    
    # split validation dataset from train set
    train_data, test_data = ds['train'], ds['test']
    train_data, val_data = random_split(train_data, [
        int(0.8 * len(train_data)),
        len(train_data) - int(0.8 * len(train_data))
    ])

    # create dataloaders
    val_loader = DataLoader(val_data, batch_size=config.trainer.batch_size, shuffle=False)
    test_loader = DataLoader(test_data, batch_size=config.trainer.batch_size, shuffle=False)
    train_loader = DataLoader(train_data, batch_size=config.trainer.batch_size, shuffle=True)
 
    # set up wandb 
    wandb_config = config.dict(exclude={"active"})
    wandb_config['data']['dataset'] = config.data.dataset_info.builder_name
    wandb_config['seed'] = args.seed
    wandb.init(config=wandb_config)

    # create the validater and tester
    validator = Evaluator(trainer.unwrapped_model)
    tester = Evaluator(trainer.unwrapped_model)
    # attach metrics
    attach_metrics(trainer, tag="train")
    attach_metrics(validator, tag="val")
    attach_metrics(tester, tag="test")
    # attach progress bar
    ProgressBar(ascii=True).attach(trainer, output_transform=lambda output: {'L': output['loss']})
    ProgressBar(ascii=True, desc='Validating').attach(validator)
    ProgressBar(ascii=True, desc='Testing').attach(tester)
    
    # add event handler for testing and logging
    @trainer.on(Events.EPOCH_COMPLETED)
    def test_and_log(engine):
        # evaluate on val and test set
        val_state = validator.run(val_loader)
        test_state = tester.run(test_loader)
        # log both train and test metrics
        print("Step:", engine.state.iteration, "-" * 15)
        print("Train F-Score:", engine.state.metrics['train/F'])
        print("Val   F-Score:", val_state.metrics['val/F'])
        print("Test  F-Score:", test_state.metrics['test/F'])
        wandb.log(
            test_state.metrics | val_state.metrics | engine.state.metrics, 
            step=engine.state.iteration
        )

    # run training
    trainer.run(train_loader, **config.trainer.run_kwargs)
   
if __name__ == '__main__':
    main() 



from __future__ import annotations
import os
# ray
import ray
from ray.tune import Tuner
from ray.air import RunConfig
from ray.air.callbacks.wandb import WandbLoggerCallback
# active
from active import helpers
from active import strategies
from active.rl import stream
from active.rl import extractors
# reward metric and evaluator (only used for output transform)
from ignite.metrics import Fbeta
from active.helpers.evaluator import Evaluator
# callbacks
from ray.rllib.algorithms.callbacks import MultiCallbacks
from active.rl.callbacks import (
    CustomMetricsFromEnvCallbacks,
    LoggingCallbacks
)
# configs
from active.scripts.run_train import Task, ModelConfig
from active.scripts.run_active import AlExperimentConfig
# others
import gym
from enum import Enum
from pydantic import BaseModel, validator
from typing import Type, List, Dict, Any, Union

def load_model_and_policy_pool_data(
    experiment:AlExperimentConfig,
    policy:FeatureExtractorConfig
):
    """ Helper function to load and preprocess pool datasets
        Note that this is non-trivial as the pool dataset has
        to be prepared for both the model and the policy's
        feature extractor while keeping the index-relationship
        between items (i.e. policy_pool[i] ~ model_pool[i]).
    """
    # get model and policy tokenizer
    model_tokenizer = experiment.model.tokenizer
    policy_tokenizer = policy.tokenizer
    # pad token ids
    model_pad_token_id = model_tokenizer.pad_token_id
    policy_pad_token_id = policy_tokenizer.pad_token_id   

    # disable filtering in data processing as this might lead to
    # differences in the model vs policy dataset
    model_data = experiment.data.copy(update={'min_sequence_length': 0})
    # create policy data config by updating only the maximum
    # sequence length but keeping minimum sequence length of zero set above
    policy_data = model_data.copy(update={'max_sequence_length': policy.max_sequence_length})

    # load and preprocess pool datasets
    model_pool = model_data.load_dataset(model_tokenizer, split={'pool': 'train'})['pool']
    policy_pool = policy_data.load_dataset(policy_tokenizer, split={'pool': 'train'})['pool']

    # undo type formatting to allow combining the datasets
    model_pool.set_format()
    policy_pool.set_format()
    # build combined dataset for filtering
    combined_pool = model_pool \
        .add_column('policy_input_ids', policy_pool['input_ids']) \
        .add_column('policy_attention_mask', policy_pool['attention_mask']) \
        .add_column('policy_labels', policy_pool['labels'])
    # use numpy arrays for filtering
    combined_pool.set_format(type='np')
    # make sure all pools are of the same size
    assert len(model_pool) == len(combined_pool) == len(policy_pool)
    
    # filter
    filter_ = lambda e: ((e['input_ids'] != model_pad_token_id).sum() > experiment.data.min_sequence_length) or \
        ((e['policy_input_ids'] != policy_pad_token_id).sum() > policy.min_sequence_length)
    combined_pool.filter(filter_, batched=False, desc='Filter')
    # split combined dataset back into model and policy datasets
    model_pool = combined_pool.remove_columns(['policy_input_ids', 'policy_attention_mask', 'policy_labels'])
    policy_pool = combined_pool.remove_columns(['input_ids', 'attention_mask', 'labels']) \
        .rename_column('policy_input_ids', 'input_ids') \
        .rename_column('policy_attention_mask', 'attention_mask') \
        .rename_column('policy_labels', 'labels')

    # set formats
    model_pool.set_format(type='torch', columns=['input_ids', 'attention_mask', 'labels'])
    policy_pool.set_format(type='torch', columns=['input_ids', 'attention_mask', 'labels'])
    # return pool datasets
    return model_pool, policy_pool
    
class RayStreamBasedEnv(stream.env.StreamBasedEnv):
    """ Stream-based environment slightly modified to be
        compatible with ray.rllib 

        Main changes to the default stream-based env are:

            (1): `__init__` expects a dictionary containing an experiment configuration
                 and a feature-extractor configuration instance
            (2): model and data loading/preparation are delayed until first
                 call of `reset` to avoid overhead of driver copy
                 (see `Expensive Environments` at https://docs.ray.io/en/latest/rllib/rllib-env.html)
    """

    def __init__(self, env_config):
        self._is_initialized = False
        # save configs for initialization in reset
        self.experiment = env_config['experiment']
        self.feature_extractor = env_config['feature_extractor']
        # dummy initialize environment, set low-resource values
        # expecially those needed for the observation space
        super(RayStreamBasedEnv, self).__init__(
            # budget will be upper bounded by pool size which is zero
            # has to be set in reset
            budget=0,
            # pass low-resource arguments
            query_size=self.experiment.active.query_size,
            metric=Fbeta(beta=1, output_transform=Evaluator.get_logits_and_labels),
            # engine and strategy are resource heavy as they need depend on the model
            engine=None,
            query_strategy=None,
            # data is obviously resource heavy :)
            policy_pool_data=[],
            model_pool_data=[],
            model_test_data=[],
            # as data is not given the layout needs to be set explicitly
            policy_sequence_length=self.feature_extractor.max_sequence_length,
            model_sequence_length=self.experiment.data.max_sequence_length,
            # TODO: use the maximum number of labels over all experiment/data setups
            max_num_labels=len(self.experiment.data.label_space)
        )

    def reset(self):

        # initialize on first call
        if not self._is_initialized:
            # get the observation space before initialization
            obs_space = self.observation_space
            # load the pool datasets
            model_pool_data, policy_pool_data = load_model_and_policy_pool_data(
                experiment=self.experiment,
                policy=self.feature_extractor
            )
            # load the model test data
            # compared to the pool data this is easy as only
            # the model needs has access to the test dataset
            model_test_data = self.experiment.load_dataset(split={'test': 'test'})['test']
            # re-initialize with actual values
            super(RayStreamBasedEnv, self).__init__(
                # now budget can be set since the
                # pool datasets are also provided
                budget=self.experiment.active.budget,
                # these are unchanged
                query_size=self.state.query_size,
                metric=self.metric,
                # build active learning engine
                engine=helpers.engine.ActiveLearningEngine(
                    trainer=self.experiment.build_trainer(),
                    trainer_run_kwargs=self.experiment.trainer.run_kwargs,
                    train_batch_size=self.experiment.trainer.batch_size,
                    eval_batch_size=self.experiment.trainer.batch_size,
                    train_val_ratio=0.9
                ),
                # set query strategy
                query_strategy=strategies.Random(),
                # pass the preprocessed datasets
                policy_pool_data=policy_pool_data,
                model_pool_data=model_pool_data,
                model_test_data=model_test_data,
                # keep these values around to make sure the
                # observation space doesn't change due to re-initialization
                policy_sequence_length=self.feature_extractor.max_sequence_length,
                model_sequence_length=self.experiment.data.max_sequence_length,
                # TODO: use the maximum number of labels over all experiment/data setups
                max_num_labels=len(self.experiment.data.label_space)
            )
            # make sure the observation spaces match
            assert obs_space == self.observation_space, "Detected change in observation space after environment initialization"
            # mark as initialized
            self._is_initialized = True

        # reset
        return super(RayStreamBasedEnv, self).reset()

class Approach(Enum):
    """ Enum defining supported active learning approaches: 
            (1) stream:
                data is presented to the strategy/agent as a stream.
                Agent decides whether to annoate or to dicard datapoints.
            (2) pool:
                data is presented to strategy/agent as pool.
                Agent selects subset to annotate next.
    """
    STREAM = 'stream'
    POOL = 'pool'

    @property
    def env_type(self) -> Type[gym.Env]:
        if self == Approach.STREAM:
            return RayStreamBasedEnv
        elif self == Approach.POOL:
            raise NotImplementedError()

class FeatureExtractor(Enum):
    """ Enum defining supported feature extractors:
            (1) transformer:
                simple transformer model as feature extractor.
                Only uses input sequence and does not depend on model output.
    """
    TRANSFORMER = 'transformer'

    @property
    def type(self) -> Type[extractors.FeatureExtractor]:
        if self is FeatureExtractor.TRANSFORMER:
            return extractors.TransformerFeatureExtractor        

class Algorithm(Enum):
    """ Enum defining supported algorithms:
            (1) dqn: Deep-Q-Network
    """
    DQN = "dqn"
    PPO = "ppo"

    @property
    def type(self) -> Type[ray.rllib.algorithms.Algorithm]:
        if self is Algorithm.DQN:
            from ray.rllib.algorithms.dqn import DQN
            return DQN
        if self is Algorithm.PPO:
            from ray.rllib.algorithms.ppo import PPO
            return PPO

class Model(Enum):
    """ Enum definig supported models:
            (1) stream/dqn-model: simple Deep-Q-Learning Model for stream-based approach
    """
    STREAM_DQN_MODEL = "stream/dqn-model"

    @property
    def type(self) -> Type[ray.rllib.models.torch.torch_modelv2.TorchModelV2]:
        if self is Model.STREAM_DQN_MODEL:
            return stream.model.DQNModel

class FeatureExtractorConfig(ModelConfig):
    """ Feature Extractor Configuration Model """
    # task must be bio tagging
    task:Task =Task.BIO_TAGGING

    # feature extractor type used by policy
    feature_extractor_type:FeatureExtractor
    # sequence length
    max_sequence_length:int
    min_sequence_length:int

    @property
    def model_config(self) -> Dict[str, Any]:
        return {
            'feature_extractor_type': self.feature_extractor_type.type,
            'feature_extractor_config': {
                'pretrained_ckpt': self.pretrained_ckpt
            }
        }

class AlgorithmConfig(BaseModel):
    """ Algorithm Configuration Model"""
    # algorithm to use and algorithm specific parameters
    algo:Algorithm
    algo_config:Dict[str, Any]
    # general algorithm config
    rollout_fragment_length:int
    train_batch_size:int

    # convergence criteria
    max_timesteps:int

    @property
    def ray_config(self) -> Dict[str, Any]:
        return {
            "rollout_fragment_length": self.rollout_fragment_length,
            "train_batch_size": self.train_batch_size,
            **self.algo_config
        }

class WorkersConfig(BaseModel):
    """ Workers Configuration Model """
    num_workers:int
    num_gpus_per_worker:float
    num_envs_per_worker:int
    remote_worker_envs:bool =False

    @property
    def ray_config(self) -> Dict[str, Any]:
        return self.dict()

def parse_experiment_config(cls, value) -> AlExperimentConfig:
    # check if value is a config instance
    if isinstance(value, AlExperimentConfig):
        return value
    # otherwise it should be a path to a valid config
    if not os.path.isfile(value):
        raise ValueError("Experiment configuration file not found: %s" % value)
    # try to load the config
    return AlExperimentConfig.parse_file(value)
    
class EvaluationConfig(BaseModel):
    """ Evaluation Configuration Model """
    # passed from RlExperiment
    _feature_extractor:FeatureExtractor =None
    # basic evaluation setup
    evaluation_interval:int
    evaluation_duration:int
    evaluation_num_workers:int
    # evaluation environemt
    env_config:Union[str, AlExperimentConfig]

    # validator to parse config
    _parse_env_config = validator('env_config', allow_reuse=True)(parse_experiment_config)

    @property
    def ray_config(self) -> Dict[str, Any]:
        return {
            # basic evaluation settings
            "evaluation_interval": self.evaluation_interval,
            "evaluation_duration": self.evaluation_duration,
            "evaluation_num_workers": self.evaluation_num_workers,
            # set evaluation environment
            "evaluation_config": {
                "env_config": {
                    "experiment": self.env_config,
                    "feature_extractor": self._feature_extractor
                }
            }
        }

class RlExperimentConfig(BaseModel):
    """ Reinforcement Learning Experiment Configuration Model"""
    # approach to use
    approach:Approach
    # algorithm and worker setup
    algorithm:AlgorithmConfig
    workers:WorkersConfig
    
    # model config
    model_type:Model
    feature_extractor:FeatureExtractorConfig
    # experiments -> environment setups
    env_configs:List[Union[str, AlExperimentConfig]]    

    # evaluation config
    evaluation:EvaluationConfig

    # validator to parse configs
    _parse_env_configs = validator('env_configs', each_item=True, allow_reuse=True)(parse_experiment_config)

    @validator('evaluation')
    def _pass_feature_extractor_to_evaluation(cls, value, values):
        assert 'feature_extractor' in values
        if isinstance(value, BaseModel):
            return value.copy(update={'_feature_extractor': values.get('feature_extractor')})
        elif isinstance(value, dict):
            return value | {'_feature_extractor': values.get('feature_extractor')}

    @property
    def ray_config(self) -> Dict[str, Any]:
        return {
            # set framework to use
            "framework": "torch",
            # specify model
            "model": {
                "custom_model": self.model_type.type,
                "custom_model_config": self.feature_extractor.model_config
            },
            # set up environments
            "env": self.approach.env_type,
            "env_config": {
                "experiment": self.env_configs[0],
                "feature_extractor": self.feature_extractor
            },
            # add configs from all sub-models
            **self.algorithm.ray_config,
            **self.workers.ray_config,
            **self.evaluation.ray_config,
        }

def main():
    from argparse import ArgumentParser
    # build argument parser
    parser = ArgumentParser()
    parser.add_argument("--config", type=str, required=True, help="Path to valid RL experiment config")
    parser.add_argument("--seed", type=int, default=1337, help="Random Seed")
    # parse arguments
    args = parser.parse_args()

    # parse configuration and apply overwrites
    config = RlExperimentConfig.parse_file(args.config)    
    print("Config:", config.json(indent=2))

    # build ray config and add callbacks
    ray_config = config.ray_config
    ray_config['callbacks'] = MultiCallbacks([
        CustomMetricsFromEnvCallbacks,
        LoggingCallbacks
    ])
    ray_config['log_level'] = "INFO"

    # connect to ray cluster
    ray.init(address='auto')

    # start training
    Tuner(
        config.algorithm.algo.type,
        param_space=ray_config,
        run_config=RunConfig(
            # stop config
            stop=dict(
                timesteps_total=config.algorithm.max_timesteps,
            ),
            # setup wandb callback
            callbacks=[
                WandbLoggerCallback(
                    project="rl-active-learning",
                    group=None,
                    log_config=False,
                    save_checkpoints=False,
                    config=config.dict()
                )
            ],
            verbose=1
        ),
    ).fit()
    # shutdown
    ray.shutdown()

if __name__ == '__main__':
    main()
